#include "keyboard.h"

#include "exceptions.h"

#include <QX11Info>
#include <X11/XKBlib.h>
#include <X11/Xatom.h>
#include <X11/keysym.h>

//#include <X11/extensions/XKBrules.h>
#define	_XKB_RF_NAMES_PROP_ATOM "_XKB_RULES_NAMES"
#define	_XKB_RF_NAMES_PROP_MAXLEN 1024

#define explicit dont_use_cxx_explicit
#include <xcb/xkb.h>
#undef explicit

typedef union {
    struct {
        uint8_t response_type;
        uint8_t xkbType;
        uint16_t sequence;
        xcb_timestamp_t time;
        uint8_t deviceID;
    } any;
    xcb_xkb_map_notify_event_t map_notify;
    xcb_xkb_state_notify_event_t state_notify;
} _xkb_event;


XKeyboard* XKeyboard::m_instance = nullptr;


XKeyboard::XKeyboard(QObject* parent) :
    QObject(parent)
{
}

XKeyboard* XKeyboard::instance()
{
    if (!m_instance) {
        m_instance = new XKeyboard();
    }

    return m_instance;
}

XKeyboard::~XKeyboard()
{
}

void XKeyboard::init()
{
    int opcode_rtrn, error_rtrn;
    int major = XkbMajorVersion;
    int minor = XkbMinorVersion;

    if (!XkbLibraryVersion(&major, &minor)) {
        throw CommonException(tr("Incompatible Xlib XKB extension v.%1.%2!=%3.%4 - try to rebuild this program")
                              .arg(major)
                              .arg(minor)
                              .arg(XkbMajorVersion)
                              .arg(XkbMinorVersion));
    }

    if (!XkbQueryExtension(QX11Info::display(), &opcode_rtrn, &m_xkbEventId, &error_rtrn, &major, &minor)) {
        throw CommonException(tr("Incompatible or disabled X server XKB extension - check X server"));
    }
}

int XKeyboard::getXkbEventId() const
{
    return m_xkbEventId;
}

const QStringList XKeyboard::getAvailableLayouts() const
{
    QStringList groups;

    const auto dpy = QX11Info::display();
    const auto rules_atom = XInternAtom(dpy, _XKB_RF_NAMES_PROP_ATOM, False);

    if (rules_atom == None) {
        throw CommonException(tr("Could not find the atom %1").arg(_XKB_RF_NAMES_PROP_ATOM));
    }

    Atom real_prop_type;
    int fmt;
    unsigned long nitems;
    unsigned long extra_bytes;
    char* prop_data = nullptr;

    const auto ret = XGetWindowProperty(dpy,
                                        QX11Info::appRootWindow(),
                                        rules_atom, 0L, _XKB_RF_NAMES_PROP_MAXLEN,
                                        False, XA_STRING, &real_prop_type, &fmt,
                                        &nitems, &extra_bytes,
                                        static_cast<unsigned char**>(static_cast<void*>(&prop_data)));

    if (ret != Success) {
        throw CommonException(tr("Could not get the property XGetWindowProperty"));
    }

    if (extra_bytes > 0 ||
        real_prop_type != XA_STRING ||
        fmt != 8)
    {
        if (prop_data) {
            XFree(prop_data);
        }
        throw CommonException(tr("Wrong property format XGetWindowProperty"));
    }

    char* out = prop_data;
    out += strlen(out) + 1;

    if (static_cast<unsigned long>(out - prop_data) < nitems) {
        out += strlen(out) + 1;
    }

    if (static_cast<unsigned long>(out - prop_data) < nitems && *out) {
        groups = QString::fromLocal8Bit(out).split(',');
    }

    XFree (prop_data);
    return groups;
}

const QString XKeyboard::getCurrentLayout() const
{
    XkbStateRec xkbStateRec;
    XkbGetState(QX11Info::display(), XkbUseCoreKbd, &xkbStateRec);

    return getAvailableLayouts().value(xkbStateRec.group);
}

const QString XKeyboard::getLayout(const uint8_t group) const
{
    return getAvailableLayouts().value(group);
}

void XKeyboard::changeLayout() const
{
    QStringList layouts = getAvailableLayouts();

    if (layouts.size() > 1) {
        XkbStateRec xkbStateRec;
        XkbGetState(QX11Info::display(), XkbUseCoreKbd, &xkbStateRec);
        int index = xkbStateRec.group + 1;

        if (index >= layouts.size()) {
            index = 0;
        }

        XkbLockGroup(QX11Info::display(), XkbUseCoreKbd, static_cast<unsigned int>(index));
    }
}

bool XKeyboard::capsLockState() const
{
    return keyState(XK_Caps_Lock);
}

void XKeyboard::xcbGenericEventHandle(const QByteArray& eventType, void* message, long* result)
{
    Q_UNUSED(result);

    if (eventType != "xcb_generic_event_t") {
        return;
    }

    const auto event = static_cast<xcb_generic_event_t*>(message);

    switch (event->response_type) {
    case XCB_KEY_PRESS:
    case XCB_KEY_RELEASE: {
        const auto keyEvent = static_cast<xcb_key_press_event_t*>(static_cast<void*>(event));
        const auto keycode = XKeysymToKeycode(QX11Info::display(), XK_Caps_Lock);
        if (keyEvent->detail == keycode && (keyEvent->state & XCB_MOD_MASK_LOCK)) {
            emit capsLockStateChanged(capsLockState());
        }
        break;
    }
    default:
        break;
    }

    unsigned int responce_type = event->response_type & ~0x80;
    if (responce_type == static_cast<unsigned int>(m_xkbEventId)) {
        const auto xkbEvent = reinterpret_cast<_xkb_event*>(event);
        if (xkbEvent && xkbEvent->any.xkbType == XCB_XKB_STATE_NOTIFY) {
            emit layoutChanged(getLayout(xkbEvent->state_notify.group));
        }
    }
}

bool XKeyboard::keyState(int iKey) const
{
    int iKeyMask = 0;
    Window wDummy1, wDummy2;
    int iDummy3, iDummy4, iDummy5, iDummy6;
    unsigned int iMask;
    const auto display = QX11Info::display();

    const auto map = XGetModifierMapping(display);
    const auto keyCode = XKeysymToKeycode(display, static_cast<KeySym>(iKey));
    if (keyCode == NoSymbol) {
        return false;
    }
    for (int i = 0; i < 8; ++i) {
        if (map->modifiermap[map->max_keypermod * i] == keyCode) {
            iKeyMask = 1 << i;
        }
    }
    XQueryPointer(display, DefaultRootWindow(display), &wDummy1, &wDummy2, &iDummy3, &iDummy4, &iDummy5, &iDummy6, &iMask);
    XFreeModifiermap(map);

    return (iMask & static_cast<unsigned int>(iKeyMask)) != 0;
}

