#include "usersmanager.h"

#include "stssettings.h"
#include "tokenwatcher.h"

#include <unistd.h>


UsersManager::UsersManager(const UsersPwdList& users,
                           const IconList& icons,
                           StsSettings* stsSettings,
                           const bool showTokenUsers,
                           const bool showUsers,
                           const QString defaultUser,
                           const PreselUser preselUser,
                           const bool userCompletion,
                           const FirstScreen firstScreen,
                           QObject* parent) :
    QObject(parent),
    Manager("users"),
    showTokenUsers(showTokenUsers),
    showUsers(showUsers),
    m_stsSettings(stsSettings),
    m_defaultUser(defaultUser),
    m_preselUser(preselUser),
    m_userCompletion(userCompletion),
    m_firstScreen(firstScreen)
{
    const auto defaultIcon = icons.value("", QIcon());
    const auto hostname = getHostname();

    TokenWatcher *watcher = TokenWatcher::instance();
    connect(watcher, &TokenWatcher::tokenUsersChanged, this, &UsersManager::tokenUsersChanged);
    QStringList tokenUserList = watcher->init();

    foreach (const auto& user, users) {
        UserData data;
        data.username = user.username;
        data.hostname = hostname;
        data.icon = icons.value(user.username, defaultIcon);

        if (tokenUserList.contains(data.username)) {
            data.isTokenUser = true;

            if (showTokenUsers) {
                m_tokenUsers << data;
            }
        }

        m_users << data;
    }
}

UsersManager::~UsersManager()
{

}

const UsersData UsersManager::users() const
{
    return showTokenUsers ? m_tokenUsers : m_users;
}

UsersManager::PreselUser UsersManager::preselUser() const
{
    return m_preselUser;
}

const QString UsersManager::prevUsername() const
{
    if (!m_stsSettings) {
        return QString();
    }

    return m_stsSettings->getPrevUsername();
}

void UsersManager::savePrevUsername(const QString& username) const
{
    if (!m_stsSettings) {
        return;
    }

    m_stsSettings->savePrevUsername(username);
}

const QString UsersManager::defaultUser() const
{
    return m_defaultUser;
}

const QString UsersManager::getHostname()
{
    char buf[256];
    buf[sizeof(buf) - 1] = '\0';
    if (gethostname(buf, sizeof(buf) - 1)) {
        return "localhost";
    } else {
        return QString::fromLocal8Bit(buf);
    }
}

bool UsersManager::userCompletion() const
{
    return m_userCompletion;
}

UsersManager::FirstScreen UsersManager::firstScreen() const
{
    return m_firstScreen;
}

void UsersManager::tokenUsersChanged(const QStringList &tokenUsersList)
{
    if (showTokenUsers) {
        m_tokenUsers.clear();
    }

    for (auto &user : m_users) {
        user.isTokenUser = tokenUsersList.contains(user.username);

        if (showTokenUsers && user.isTokenUser) {
            m_tokenUsers << user;
        }
    }

    if (showTokenUsers) {
        emit usersChanged(m_tokenUsers);
    }
    else {
        emit usersChanged(m_users);
    }
}
