#include "userinfo.h"

#include <QDir>
#include <QMap>

#include <pwd.h>
#include <grp.h>


const PasswdLine PasswdLine::fromPasswd(passwd *pwd)
{
    PasswdLine line;
    line.username = QString::fromUtf8(pwd->pw_name);
    line.password = QString::fromUtf8(pwd->pw_passwd);
    line.uid = pwd->pw_uid;
    line.gid = pwd->pw_gid;
    line.gecos = QString::fromUtf8(pwd->pw_gecos);
    line.homeDir = QString::fromUtf8(pwd->pw_dir);
    line.shell = QString::fromUtf8(pwd->pw_shell);

    return line;
}



UserList::UserList(const QStringList& members)
{
    foreach (const auto& member, members) {
        if (member.startsWith('@')) {
            const auto groupname = QStringRef(&member, 1, member.count() - 1).toString();
            const auto grp = getgrnam(groupname.toUtf8());
            m_groups.insert(grp->gr_gid);
            for (; *grp->gr_mem; grp->gr_mem++) {
                m_users.insert(QString::fromUtf8(*grp->gr_mem));
            }
        } else {
            m_users.insert(member);
        }
    }
}

bool UserList::hasUser(const QString& username) const
{
    return m_users.contains(username);
}

bool UserList::hasGroup(const gid_t gid) const
{
    return m_groups.contains(gid);
}

bool UserList::hasGroups() const
{
    return !m_groups.isEmpty();
}


const UsersPwdList UserInfo::getUserList()
{
    UsersPwdList users;
    passwd* pwd;

    setpwent();
    while((pwd = getpwent()) != nullptr) {
        users << PasswdLine::fromPasswd(pwd);
    }
    endpwent();

    return users;
}

const UsersPwdList UserInfo::getUserList(const UserInfo::Filter& filter)
{
    UsersPwdList users;

    const UserList filterUsers(filter.users);
    if (!filter.showUsers && !filterUsers.hasGroups()) {
        foreach (const auto& user, filter.users) {
            const auto ps = getpwnam(user.toUtf8());
            if (!ps) {
                continue;
            }
            const auto line = PasswdLine::fromPasswd(ps);
            if (!line.uid && !filter.showRoot) {
                continue;
            }
            users << line;
        }
    } else {
        QSet<QString> usernames;
        const UserList filterNoUsers(filter.noUsers);
        foreach (const auto& user, getUserList()) {
            if (usernames.contains(user.username)) {
                continue;
            }
            if (user.uid < filter.lowUserId && (user.uid || !filter.showRoot)) {
                continue;
            }
            if (user.uid > filter.highUserId) {
                continue;
            }

            if (filter.showUsers) {
                if (filterNoUsers.hasUser(user.username)) {
                    continue;
                }
                if (filterNoUsers.hasGroup(user.gid)) {
                    continue;
                }
            } else {
                if (!filterUsers.hasUser(user.username) && !filterUsers.hasGroup(user.gid)) {
                    continue;
                }
            }

            users << user;
            usernames << user.username;
        }
    }

    if (filter.sortUsers) {
        std::sort(users.begin(), users.end(), [](const PasswdLine& first, const PasswdLine& second){
            return first.username < second.username;
        });
    }

    return users;
}


const IconList UserIcon::getUserIcons(const QString& iconsPath)
{
    IconList icons;

    const QString defaultIconPath(iconsPath + '/' + ".default.face.icon");
    if (QFile(defaultIconPath).exists()) {
        icons[""] = QIcon(defaultIconPath);
    }

    foreach (const auto& info, QDir(iconsPath).entryInfoList(QDir::Files)) {
        if (info.completeSuffix() == "face.icon") {
            icons[info.baseName()] = QIcon(info.absoluteFilePath());
        }
    }

    return icons;
}
