#include "utils.h"

#include <sys/stat.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>

#include <QLocale>
#include <QBitmap>
#include <QPainter>
#include <QScreen>

#include "exceptions.h"
#include <fly/flyicon.h>
#include <fly/flymisc.h>
#include <fly/flyfileutils.h>
#include <fly/flystring.h>


const QString Utils::shortSystemLocale()
{
    const auto localeName = QLocale::system().name();
    return  QStringRef(&localeName, 0, 2).toString();
}

QFile* Utils::openFile(const QString& path)
{
    const auto file = new QFile(path);

    if (!file->exists()) {
        delete file;
        throw FileNotExistsException(path);
    }

    if (!file->open(QFile::ReadOnly)) {
        const auto error = file->errorString();
        delete file;
        throw OpenFileException(path, QFile::ReadOnly, error);
    }

    return file;
}

const QPixmap Utils::circlePixmapFromIcon(const QIcon& icon, const QSize& size)
{
    QSize originalSize(size);
    if (!icon.availableSizes().isEmpty()) {
        originalSize = icon.availableSizes().first();
    }
    auto pixmapSize = originalSize.scaled(size, Qt::KeepAspectRatio);
    if (pixmapSize.width() < size.width()) {
        pixmapSize *= double(size.width()) / pixmapSize.width();
    } else if (pixmapSize.height() < size.height()) {
        pixmapSize *= double(size.height()) / pixmapSize.height();
    }

    auto pixmap = icon.pixmap(pixmapSize).scaled(pixmapSize, Qt::KeepAspectRatio, Qt::SmoothTransformation);

    const auto paddingSize = pixmap.size() - size;
    const QPoint padding(paddingSize.width() / 2, paddingSize.height() / 2);

    pixmap = pixmap.copy(QRect(padding, size));

    QBitmap mask(pixmap.size());
    mask.fill(Qt::white);

    QPainter painter(&mask);
    painter.setBrush(Qt::black);
    const auto radius = std::min(mask.width(), mask.height()) / 2 - 5;
    painter.drawEllipse(QPoint(mask.width() / 2, mask.height() / 2), radius, radius);

    pixmap.setMask(mask);

    return pixmap;
}

const QSize Utils::caclulateRealPaintDeviceSize(QPaintDevice* device)
{
    const auto width = double(device->width());
    const auto height = double(device->height());

    const auto wFactor = 96.0 / device->physicalDpiX();
    const auto hFactor = 96.0 / device->physicalDpiY();

    const auto subFactor = 5;

    return QSize (int(subFactor * wFactor * width), int(subFactor * hFactor * height));
}

const QPixmap Utils::circlePixmapFromIconWithEmblem(const QIcon &icon, const QSize &size, const QIcon &emblem)
{
    QIcon newIcon;
    newIcon.addPixmap(circlePixmapFromIcon(icon, size));

    return FlyIcon::withOverlay(newIcon, Qt::BottomRightCorner, emblem.pixmap(size.width() / 3, size.height() / 3)).pixmap(size);
}

std::string Utils::FindFlyWallPaper(int screenWidth, int screenHeight, bool* exists, const QString& dataDir)
{
    std::string name = "fly-default";
    std::string wpdir = Fly::concat(dataDir.toStdString(),"/wallpapers");
    const char * pathToFind[1]; //, *convertorFmt="fly-svg2png.sh %d %d %s %s";
//    pathToFind[0]=wpdir.c_str();
    pathToFind[0]="/usr/share/wallpapers/fly-dm";

    if (exists) *exists=true;

    if (!fileExists(pathToFind[0],false))
    {
        if (mkdir(pathToFind[0],S_IRWXU) != 0) fprintf(stderr,"fly-qdm: failed to create dir '%s'\n", pathToFind[0]);
        sync();
    }

    int w, h;
    std::string ex=".png", path;
    int ratioWidth, ratioHeight;
    std::tie(ratioWidth, ratioHeight) = wallpaperSizeRatio(screenWidth, screenHeight);

    for (unsigned int i=0;i<sizeof(pathToFind)/sizeof(char *);i++)
    {
        w=screenWidth; h=screenHeight; //check for screen size WxH

        path = Fly::concat(pathToFind[i], '/', w, 'x', h, '/', name, ex);
        if (fileExists(path,true)) goto found;
        else fprintf(stderr, "fly-qdm: Can't find default wallpaper '%s' for screen %dx%d, attempt to look for other size\n", path.c_str(), w, h);

        w = ratioWidth; h = ratioHeight; //check for ratio W:H

        path = Fly::concat(pathToFind[i], '/', w, ':', h, '/', name, ex);
        if (fileExists(path,true)) goto found;
        else fprintf(stderr, "fly-qdm: Can't find default wallpaper '%s' for ratio %d:%d\n", path.c_str(), w, h);

        if (!ex.empty() && ex != ".svg")
        {
            path = Fly::concat(pathToFind[i], '/', w, ':', h, '/', name, ".svg");
            if (fileExists(path,true)) {
                goto found;
//                std::string newPath = Fly::concat(dataDir.toStdString(), "/wallpapers/", screenWidth, 'x', screenHeight);

//                if (!fileExists(newPath,true)) {
//                    if (mkdir(newPath.c_str(),S_IRWXU) != 0) fprintf(stderr,"fly-qdm: failed to create dir %s\n", newPath.c_str());
//                    sync();
//                }

//                newPath += '/'; newPath += name; newPath += ".png";

//                char cmd[PATH_MAX*2];
//                int len = snprintf(cmd,sizeof(cmd)-1,convertorFmt,screenWidth,screenHeight,path.c_str(),newPath.c_str()); cmd[len]='\0';
//                fprintf(stderr,"fly-qdm: SVG conversion by: %s... in progress\n",cmd);

//                if (exists) *exists=false;
//                system(cmd); //we need the result immediately

//                path = std::move(newPath);
//                if (fileExists(path,true)) goto found;
//                else {
//                        fprintf(stderr,"fly-qdm: Can't find converted wallpaper '%s'\n", path.c_str());
//                        return std::string();
//                }
            } else {
                fprintf(stderr,"fly-qdm: Can't find default wallpaper '%s' for ratio %d:%d\n", path.c_str(), w, h);
            }
        }
    } //endfor
    return std::string();

found:
    return path;
}

std::string Utils::FlyDefaultLogoPath(std::string logoName, int desktopWidth, int desktopHeight, std::string logoColor)
{
    struct stat st;
    std::string logoPath;
    std::string logoSize; //logo standard sizes is 800, 1280, 1920, 2048 - proposed by designer Dm. Scherbakov
         if (desktopWidth<=800)  logoSize="800";
    else if (desktopWidth<=1280) logoSize="1280";
    else if (desktopWidth<=1920) logoSize="1920";
    else                         logoSize="2048";

    logoPath = logoName + "_"+logoSize + "_"+logoColor + ".png"; //1: full spec name_size_color.png
    if (stat(logoPath.c_str(),&st)==0) return logoPath;

    logoPath = logoName + "_"+logoSize + ".png"; //2: name_size.png
    if (stat(logoPath.c_str(),&st)==0) return logoPath;

    logoPath = logoName + "_"+logoColor + ".png"; //3: name_color.png
    if (stat(logoPath.c_str(),&st)==0) return logoPath;

    logoPath = logoName + ".png"; //4: name.png
    if (stat(logoPath.c_str(),&st)==0) return logoPath;

    return logoPath;
}

const QRect Utils::screenGeometry()
{
    auto *screen = QGuiApplication::primaryScreen();
    const auto geometry = screen->geometry();
    return QRect(geometry.topLeft(), geometry.size() * screen->devicePixelRatio());
}

const QImage Utils::blurred(const QImage& image, const QRect& rect, int radius, bool alphaOnly)
{
    int tab[] = { 14, 10, 8, 6, 5, 5, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2 };
    int alpha = (radius < 1)  ? 16 : (radius > 17) ? 1 : tab[radius-1];

    auto result = image.convertToFormat(QImage::Format_ARGB32_Premultiplied);
    int r1 = rect.top();
    int r2 = rect.bottom();
    int c1 = rect.left();
    int c2 = rect.right();

    int bpl = result.bytesPerLine();
    int rgba[4];
    unsigned char* p;

    int i1 = 0;
    int i2 = 3;

    if (alphaOnly)
        i1 = i2 = (QSysInfo::ByteOrder == QSysInfo::BigEndian ? 0 : 3);

    for (int col = c1; col <= c2; col++) {
        p = result.scanLine(r1) + col * 4;
        for (int i = i1; i <= i2; i++)
            rgba[i] = p[i] << 4;

        p += bpl;
        for (int j = r1; j < r2; j++, p += bpl)
            for (int i = i1; i <= i2; i++)
                p[i] = (rgba[i] += ((p[i] << 4) - rgba[i]) * alpha / 16) >> 4;
    }

    for (int row = r1; row <= r2; row++) {
        p = result.scanLine(row) + c1 * 4;
        for (int i = i1; i <= i2; i++)
            rgba[i] = p[i] << 4;

        p += 4;
        for (int j = c1; j < c2; j++, p += 4)
            for (int i = i1; i <= i2; i++)
                p[i] = (rgba[i] += ((p[i] << 4) - rgba[i]) * alpha / 16) >> 4;
    }

    for (int col = c1; col <= c2; col++) {
        p = result.scanLine(r2) + col * 4;
        for (int i = i1; i <= i2; i++)
            rgba[i] = p[i] << 4;

        p -= bpl;
        for (int j = r1; j < r2; j++, p -= bpl)
            for (int i = i1; i <= i2; i++)
                p[i] = (rgba[i] += ((p[i] << 4) - rgba[i]) * alpha / 16) >> 4;
    }

    for (int row = r1; row <= r2; row++) {
        p = result.scanLine(row) + c2 * 4;
        for (int i = i1; i <= i2; i++)
            rgba[i] = p[i] << 4;

        p -= 4;
        for (int j = c1; j < c2; j++, p -= 4)
            for (int i = i1; i <= i2; i++)
                p[i] = (rgba[i] += ((p[i] << 4) - rgba[i]) * alpha / 16) >> 4;
    }

    return result;
}

// Copied from libflycore
bool Utils::isUserNameInGroup(const char *groupName, const char *userName)
{
    bool res = false;
    struct passwd *pw;
    pw = getpwnam(userName); /* Fetch passwd structure (contains first group ID for user) */
    if (pw)
    {
        int ngroups=256;
        gid_t *groups;

        groups = (gid_t *)malloc(ngroups * sizeof (gid_t));
        if (groups == NULL) goto memerr;

        if (getgrouplist(userName, pw->pw_gid, groups, &ngroups) == -1) {
            fprintf(stderr, "libflycore: getgrouplist(%s,...) returned -1; ngroups = %d\n",userName,ngroups);
        }
        else {
            for (int j = 0; j < ngroups; j++) {
                struct group *gr = getgrgid(groups[j]);
                if (gr && gr->gr_name && strcmp(gr->gr_name,groupName)==0) {
                    //fprintf(stderr,"user %s in group %s\n",userName.c_str(),groupName);
                    res=true; break;
                }
            }
        }
        free(groups);
    }
    else fprintf(stderr,"libflycore: getpwnam(%s) failed\n",userName);

    return res;
memerr:
    fprintf(stderr,"libflycore: isUserInGroup() memory allocation error\n");
    return res;
}
