#include "loginwidget.h"

#include "utils.h"
#include "themewidgetcreator.h"
#include "usercellwidget.h"
#include "swipegesturerecognizer.h"

#include <QGridLayout>
#include <QHeaderView>
#include <QShortcut>
#include <QTimer>
#include <QCompleter>
#include <QStringListModel>
#include <QDebug>


LoginWidget::LoginWidget(QWidget* parent) :
    QWidget(parent),
    m_themeDescription(""), // TODO
    m_verticalUserWidgetFile(nullptr),
    m_horizontalUserWidgetFile(nullptr),
    m_shortUsersColumns(0),
    m_currentUsersPage(0)
{
}

LoginWidget::~LoginWidget()
{
    if (m_verticalUserWidgetFile) {
        delete m_verticalUserWidgetFile;
    }

    if (m_horizontalUserWidgetFile) {
        delete m_horizontalUserWidgetFile;
    }
}

void LoginWidget::load(QWidget* sourceWidget, const ThemeDescription& description)
{
    m_themeDescription = description;

    const auto themePath = m_themeDescription.dirPath();

    m_verticalUserWidgetFile = Utils::openFile(themePath + "/vertical_user_item.ui");
    m_horizontalUserWidgetFile = Utils::openFile(themePath + "/horizontal_user_item.ui");
    m_whiteCircleIcon = QIcon(themePath + "/icon_white_circle.svg");
    m_emptyWhiteCircleIcon = QIcon(themePath + "/icon_empty_white_circle.svg");
    m_insertTokenIcon = QIcon(themePath + "/icon-token.png");

    const auto layout = new QGridLayout();
    layout->setMargin(0);
    layout->addWidget(sourceWidget);
    setLayout(layout);

    m_userIconLabel = Utils::findWidget<QLabel*>("selectedUserIconLabel", this);
    m_usernameLabel = Utils::findWidget<QLabel*>("usernameLabel", this);
    m_userHostnameLabel = Utils::findWidget<QLabel*>("userHostnameLabel", this);
    m_shortUsersTableWidget = Utils::findWidget<QTableWidget*>("shortUsersTableWidget", this);
    m_allUsersTableWidget = Utils::findWidget<QTableWidget*>("allUsersTableWidget", this);
    m_userPagesTableWidget = Utils::findWidget<QTableWidget*>("userPagesTableWidget", this);
    m_usersStackedWidget = Utils::findWidget<QStackedWidget*>("usersStackedWidget", this);
    m_commonStackedWidget = Utils::findWidget<QStackedWidget*>("commonStackedWidget", this);
    m_changeUserToolButton = Utils::findWidget<QToolButton*>("changeUserToolButton", this);
    m_usernameLineEdit = Utils::findWidget<QLineEdit*>("usernameLineEdit", this);
    m_passwordUsernameLineEdit = Utils::findWidget<QLineEdit*>("passwordUsernameLineEdit", this);
    m_loginUsernameToolButton = Utils::findWidget<QToolButton*>("loginUsernameToolButton", this);
    m_hostsComboBox = Utils::findWidget<QComboBox*>("hostsComboBox", this);
    m_passwordLineEdit = Utils::findWidget<QLineEdit*>("passwordLineEdit", this);
    m_loginToolButton = Utils::findWidget<QToolButton*>("loginToolButton", this);
    m_loginStackedWidget = Utils::findWidget<QStackedWidget*>("loginStackedWidget", this);
    m_userSelectionWidget = Utils::findWidget<QWidget*>("userSelectionWidget", this);
    m_capsLockLabel = Utils::findWidget<QLabel*>("capsLockLabel", this);
    m_capsLockUsernameLabel = Utils::findWidget<QLabel*>("capsLockUsernameLabel", this);
    m_usersFilterLineEdit = Utils::findWidget<QLineEdit*>("usersFilterLineEdit", this);
    m_hostnameLabel = Utils::findWidget<QLabel *>("hostnameLabel", this);

    m_shortUsersColumns = m_shortUsersTableWidget->columnCount();
    m_allUsersRows = m_allUsersTableWidget->rowCount();
    m_allUsersColumns = m_allUsersTableWidget->columnCount();
    m_allUsersDefaultSectionSize = m_allUsersTableWidget->horizontalHeader()->defaultSectionSize();

    setCurrentUsersWidgetPage(UsersWidgetPage::ChangeUserButton);

    connect(m_loginToolButton, &QToolButton::clicked,
            this, &LoginWidget::loginToolButtonClicked);

    connect(m_loginUsernameToolButton, &QToolButton::clicked,
            this, &LoginWidget::loginUsernameToolButtonClicked);

    connect(m_changeUserToolButton, &QToolButton::clicked,
            this, &LoginWidget::changeUserToolButtonClicked);

    connect(m_userPagesTableWidget, &QTableWidget::cellPressed,
            this, &LoginWidget::userPagesWidgetPressed);
    connect(m_userPagesTableWidget, &QTableWidget::cellActivated,
            this, &LoginWidget::userPagesWidgetActivated);

    connect(m_allUsersTableWidget, &QTableWidget::cellClicked,
            this, &LoginWidget::allUsersTableWidgetPressed);
    connect(m_allUsersTableWidget, &QTableWidget::cellActivated,
            this, &LoginWidget::allUsersTableWidgetPressed);

    connect(m_shortUsersTableWidget, &QTableWidget::cellPressed,
            this, &LoginWidget::shortUsersTableWidgetPressed);
    connect(m_shortUsersTableWidget, &QTableWidget::cellActivated,
            this, &LoginWidget::shortUsersTableWidgetPressed);

    const auto changeSelectingUserShortcut = new QShortcut(this);
    connect(changeSelectingUserShortcut, &QShortcut::activated,
            this, &LoginWidget::changeSelectingUserShortcutActivated);

    const auto escapeKeyShortcut = new QShortcut(this);
    escapeKeyShortcut->setKey(Qt::Key_Escape);
    connect(escapeKeyShortcut, &QShortcut::activated,
            this, &LoginWidget::escapeKeyShortcutActivated);

    m_passwordLineEdit->setContextMenuPolicy(Qt::ContextMenuPolicy::NoContextMenu);
    connect(m_passwordLineEdit, &QLineEdit::returnPressed,
            this, &LoginWidget::passwordLineEditReturnPressed);

    m_usernameLineEdit->setContextMenuPolicy(Qt::ContextMenuPolicy::NoContextMenu);
    connect(m_usernameLineEdit, &QLineEdit::returnPressed,
            this, &LoginWidget::usernameLineEditReturnPressed);
    m_passwordUsernameLineEdit->setContextMenuPolicy(Qt::ContextMenuPolicy::NoContextMenu);
    connect(m_passwordUsernameLineEdit, &QLineEdit::returnPressed,
            this, &LoginWidget::passwordUsernameLineEditReturnPressed);

    enableCapsLock(false);

    connect(m_usersFilterLineEdit, &QLineEdit::textEdited,
            this, &LoginWidget::filterChanged);

    setCurrentCommonWidgetPage(CommonWidgetPage::Login);
    setUsers(UsersData(), false);

    QGestureRecognizer::registerRecognizer(new SwipeGestureRecognizer());
    m_allUsersTableWidget->grabGesture(Qt::SwipeGesture);

    qApp->installEventFilter(this);

    setCurrentUserData(UserData());
    setCurrentAuthPage(AuthPage::TextEnterUsername);
}

void LoginWidget::loadGeneric(QWidget *sourceWidget, const ThemeDescription &description)
{
    m_themeDescription = description;

    const auto themePath = m_themeDescription.dirPath();

//    m_verticalUserWidgetFile = Utils::openFile(themePath + "/vertical_user_item.ui");
//    m_horizontalUserWidgetFile = Utils::openFile(themePath + "/horizontal_user_item.ui");
//    m_whiteCircleIcon = QIcon(themePath + "/icon_white_circle.svg");
//    m_emptyWhiteCircleIcon = QIcon(themePath + "/icon_empty_white_circle.svg");
//    m_insertTokenIcon = QIcon(themePath + "/icon-token.png");

    const auto layout = new QGridLayout();
    layout->setMargin(0);
    layout->addWidget(sourceWidget);
    setLayout(layout);

    m_userIconLabel = Utils::findWidget<QLabel*>("selectedUserIconLabel", this);
    m_usernameLabel = Utils::findWidget<QLabel*>("usernameLabel", this);
    m_userHostnameLabel = Utils::findWidget<QLabel*>("userHostnameLabel", this);
    m_shortUsersTableWidget = Utils::findWidget<QTableWidget*>("shortUsersTableWidget", this);
    m_allUsersTableWidget = Utils::findWidget<QTableWidget*>("allUsersTableWidget", this);
    m_userPagesTableWidget = Utils::findWidget<QTableWidget*>("userPagesTableWidget", this);
    m_usersStackedWidget = Utils::findWidget<QStackedWidget*>("usersStackedWidget", this);
    m_commonStackedWidget = Utils::findWidget<QStackedWidget*>("commonStackedWidget", this);
    m_changeUserToolButton = Utils::findWidget<QToolButton*>("changeUserToolButton", this);
    m_usernameLineEdit = Utils::findWidget<QLineEdit*>("usernameLineEdit", this);
    m_passwordUsernameLineEdit = Utils::findWidget<QLineEdit*>("passwordUsernameLineEdit", this);
    m_loginUsernameToolButton = Utils::findWidget<QToolButton*>("loginUsernameToolButton", this);
    m_hostsComboBox = Utils::findWidget<QComboBox*>("hostsComboBox", this);
    m_passwordLineEdit = Utils::findWidget<QLineEdit*>("passwordLineEdit", this);
    m_loginToolButton = Utils::findWidget<QToolButton*>("loginToolButton", this);
    m_loginStackedWidget = Utils::findWidget<QStackedWidget*>("loginStackedWidget", this);
    m_userSelectionWidget = Utils::findWidget<QWidget*>("userSelectionWidget", this);
    m_capsLockLabel = Utils::findWidget<QLabel*>("capsLockLabel", this);
    m_capsLockUsernameLabel = Utils::findWidget<QLabel*>("capsLockUsernameLabel", this);
    m_usersFilterLineEdit = Utils::findWidget<QLineEdit*>("usersFilterLineEdit", this);
    m_hostnameLabel = Utils::findWidget<QLabel *>("hostnameLabel", this);

    m_shortUsersColumns = m_shortUsersTableWidget->columnCount();
    m_allUsersRows = m_allUsersTableWidget->rowCount();
    m_allUsersColumns = m_allUsersTableWidget->columnCount();
    m_allUsersDefaultSectionSize = m_allUsersTableWidget->horizontalHeader()->defaultSectionSize();

    setCurrentUsersWidgetPage(UsersWidgetPage::ChangeUserButton);

    connect(m_loginToolButton, &QToolButton::clicked,
            this, &LoginWidget::genericToolButtonClicked);

    m_passwordLineEdit->setContextMenuPolicy(Qt::ContextMenuPolicy::NoContextMenu);
    connect(m_passwordLineEdit, &QLineEdit::returnPressed,
            this, &LoginWidget::genericToolButtonClicked);

    enableCapsLock(false);

    setCurrentCommonWidgetPage(CommonWidgetPage::Login);
    setUsers(UsersData(), false);

    QGestureRecognizer::registerRecognizer(new SwipeGestureRecognizer());
    m_allUsersTableWidget->grabGesture(Qt::SwipeGesture);

    qApp->installEventFilter(this);

    setCurrentUserData(UserData());
    setCurrentAuthPage(AuthPage::ShortCutsEnterUsername);
}

void LoginWidget::enableUserSelection(const bool enable)
{
    m_isUserSelectionEnabled = enable;
    m_userSelectionWidget->setVisible(m_usersData.count() > 1 && (m_isUserSelectionEnabled || currentAuthPage() == AuthPage::ShortCutsEnterUsername));
}

void LoginWidget::enableLoginWidgets(const bool enable)
{
    m_passwordLineEdit->setEnabled(enable);
    m_loginToolButton->setEnabled(enable);
    m_changeUserToolButton->setEnabled(enable);

    m_hostsComboBox->setEnabled(enable);
    m_usernameLineEdit->setEnabled(enable);
    m_passwordUsernameLineEdit->setEnabled(enable);
    m_loginUsernameToolButton->setEnabled(enable);
}

void LoginWidget::enableHostnameSelection(const bool enable)
{
    m_hostsComboBox->setVisible(enable);
}

void LoginWidget::enableCapsLock(const bool enable)
{
    m_capsLockLabel->setVisible(enable);
    m_capsLockUsernameLabel->setVisible(enable);
}

void LoginWidget::enableUsernameCompleter(const bool enable)
{
    const auto oldCompletor = m_usernameLineEdit->completer();
    if (enable && !oldCompletor) {
        const auto newCompletion = new QCompleter(this);
        newCompletion->setCompletionMode(QCompleter::InlineCompletion);
        m_usernameLineEdit->setCompleter(newCompletion);
    } else if (!enable && oldCompletor) {
        oldCompletor->deleteLater();
    }

    updateUsernameCompleterValues();
}

void LoginWidget::enableTokenMessage(const bool enable)
{
    if (enable) {
        m_usernameLabel->setText(m_insertTokenMessage);

        const auto labelSize = Utils::caclulateRealPaintDeviceSize(m_userIconLabel);
        m_userIconLabel->setPixmap(Utils::circlePixmapFromIcon(m_insertTokenIcon, labelSize));
    }
    m_userHostnameLabel->setVisible(!enable);
    m_passwordLineEdit->clear();
    m_passwordLineEdit->setVisible(!enable);
    m_loginToolButton->setVisible(!enable);
}

void LoginWidget::clearPassword()
{
    m_passwordLineEdit->clear();
    m_passwordUsernameLineEdit->clear();
}

const QString LoginWidget::currentUsername() const
{
    switch (currentAuthPage()) {
    case AuthPage::ShortCutsEnterUsername:
        return m_usernameLabel->text();
    case AuthPage::TextEnterUsername:
        return m_usernameLineEdit->text();
    default:
        return QString();
    }
}

const QString LoginWidget::currentPassword() const
{
    switch (currentAuthPage()) {
    case AuthPage::ShortCutsEnterUsername:
        return m_passwordLineEdit->text();
    case AuthPage::TextEnterUsername:
        return m_passwordUsernameLineEdit->text();
    default:
        return QString();
    }
}

void LoginWidget::setFocusOnPassword()
{
    switch (currentAuthPage()) {
    case AuthPage::ShortCutsEnterUsername:
        m_passwordLineEdit->setFocus();
        break;
    case AuthPage::TextEnterUsername:
        m_passwordUsernameLineEdit->setFocus();
        break;
    default:
        break;
    }
}

void LoginWidget::setUsers(const UsersData& users, const bool &showTokenUsers)
{
    m_usersData = users;

    m_showTokenUsers = showTokenUsers;

    if (!m_showTokenUsers && users.count() == 0) {
        m_userSelectionWidget->setVisible(false);
        return;
    }

    m_userSelectionWidget->setVisible(users.count() > 1 && (m_isUserSelectionEnabled || currentAuthPage() == AuthPage::ShortCutsEnterUsername));

    setUserLists(users);

    if (showTokenUsers) {
        enableTokenMessage(true);
    }
}

void LoginWidget::setCurrentUser(const QString& username)
{
    for (const auto& data : m_usersData) {
        if (data.username == username) {
            setCurrentUserData(data);
            setCurrentAuthPage(AuthPage::ShortCutsEnterUsername);
            return;
        }
    }

    setCurrentUserData(UserData());
    setCurrentAuthPage(AuthPage::TextEnterUsername);
    m_usernameLineEdit->setText(username);
}

LoginWidget::AuthPage LoginWidget::currentAuthPage() const
{
    if (m_loginStackedWidget->currentWidget()->objectName() == "userInfoPage") {
        return AuthPage::ShortCutsEnterUsername;
    } else if (m_loginStackedWidget->currentWidget()->objectName() == "enterUsernamePage") {
        return AuthPage::TextEnterUsername;
    } else {
        return AuthPage::UnknownPage;
    }
}

void LoginWidget::setCurrentAuthPage(const LoginWidget::AuthPage page, const bool isManual)
{
    switch (page) {
    case AuthPage::ShortCutsEnterUsername:
        setLoginStackedWidgetPage("userInfoPage");

        QTimer::singleShot(0, [&]() { // TODO
            m_passwordLineEdit->setFocus();
        });
        break;
    case AuthPage::TextEnterUsername:
        if (m_usernameLabel->text() != m_insertTokenMessage) {
            m_usernameLineEdit->setText(m_usernameLabel->text());
        }

        setLoginStackedWidgetPage("enterUsernamePage");

        if (isManual) {
            QTimer::singleShot(0, [&]() {
                m_usernameLineEdit->setFocus();
                m_usernameLineEdit->selectAll();
            });
        }
        else {
            QTimer::singleShot(0, [&]() {
                if (m_usernameLineEdit->text().isEmpty() || !m_focusPassword) {
                    m_usernameLineEdit->setFocus();
                }
                else {
                    m_passwordUsernameLineEdit->setFocus();
                }
            });
        }

        break;
    default:
        return;
    }

    m_userSelectionWidget->setVisible(m_usersData.count() > 1 && (m_isUserSelectionEnabled || currentAuthPage() == AuthPage::ShortCutsEnterUsername));

    m_passwordLineEdit->clear();
    m_passwordUsernameLineEdit->clear();
}

void LoginWidget::setNoPass(const bool enable, const QStringList &users)
{
    m_noPassLogin = enable;
    m_noPassUsers = users;
}

void LoginWidget::setHostname(const QString &hostname)
{
    if (!hostname.isEmpty()) {
        m_hostnameLabel->setText(tr("Welcome to ") + hostname);
    }
}

bool LoginWidget::eventFilter(QObject* watched, QEvent* event)
{
    const auto eventType = event->type();

    if (eventType == QEvent::Gesture) {
        if (const auto gesture = static_cast<QGestureEvent*>(event)) {
            if (const auto widget = static_cast<QWidget*>(watched)) {
                gesture->setWidget(widget);
            }
            return gestureEvent(gesture);
        }
    } else if (eventType == QEvent::MouseButtonRelease) {
        if (const auto mouseEvent = static_cast<QMouseEvent*>(event)) {

            if (currentUsersWidgetPage() == UsersWidgetPage::ShortCutUsers &&
                !m_shortUsersTableWidget->underMouse())
            {
                setCurrentUsersWidgetPage(UsersWidgetPage::ChangeUserButton);
            }

            if (currentCommonWidgetPage() == CommonWidgetPage::allUsers
                && !currentCommonWidget()->underMouse())
            {
                setCurrentCommonWidgetPage(CommonWidgetPage::Login);
            }
        }
    }

    return QWidget::eventFilter(watched, event);
}

void LoginWidget::resizeEvent(QResizeEvent* event)
{
    resizeAllUsersWidgets();
    QWidget::resizeEvent(event);
}

void LoginWidget::loginToolButtonClicked()
{
    AuthData data;
    data.username = m_usernameLabel->text();
    data.password = m_passwordLineEdit->text();
    emit loginRequest(data, QString());
}

void LoginWidget::genericToolButtonClicked()
{
    m_usernameLabel->clear();
    m_userHostnameLabel->clear();
    emit submitAnswerToPam(m_passwordLineEdit->text());
    m_passwordLineEdit->clear();
}

void LoginWidget::loginUsernameToolButtonClicked()
{
    AuthData data;
    data.username = m_usernameLineEdit->text();
    data.password = m_passwordUsernameLineEdit->text();

    if (m_hostsComboBox->count() > 0) {
        emit loginRequest(data, m_hostsComboBox->itemText(m_hostsComboBox->currentIndex()));
    }
    else {
        emit loginRequest(data, QString());
    }

}

void LoginWidget::changeUserToolButtonClicked()
{
    if (m_shortUsersTableWidget->columnCount() > 0) { // TODO
        setCurrentUsersWidgetPage(UsersWidgetPage::ShortCutUsers);
    } else {
        setCurrentCommonWidgetPage(CommonWidgetPage::allUsers);
    }
}

void LoginWidget::userPagesWidgetPressed(const int row, const int column)
{
    Q_UNUSED(row)

    m_userPagesTableWidget->clearFocus();

    setUsersPage(column);
}

void LoginWidget::userPagesWidgetActivated(const int row, const int column)
{
    Q_UNUSED(row)

    setUsersPage(column);
}

void LoginWidget::passwordLineEditReturnPressed()
{
    m_loginToolButton->click();
}

void LoginWidget::usernameLineEditReturnPressed()
{
    if (m_usernameLineEdit->text().isEmpty()) {
        return;
    }

    if (m_noPassLogin && checkUserInNoPassList(m_usernameLineEdit->text())) {
        m_loginUsernameToolButton->click();
    }
    else {
        m_passwordUsernameLineEdit->setFocus(Qt::TabFocusReason);
    }
}

void LoginWidget::passwordUsernameLineEditReturnPressed()
{
    m_loginUsernameToolButton->click();
}

void LoginWidget::allUsersTableWidgetPressed(int row, int column)
{
    const auto widget = qobject_cast<UserCellWidget*>(m_allUsersTableWidget->cellWidget(row, column));
    if (!widget) {
        return;
    }

    m_allUsersTableWidget->clearFocus();

    setCurrentUserData(widget->userData());

    setCurrentCommonWidgetPage(CommonWidgetPage::Login);
}

void LoginWidget::shortUsersTableWidgetPressed(int row, int column)
{
    const auto widget = qobject_cast<UserCellWidget*>(m_shortUsersTableWidget->cellWidget(row, column));
    if (!widget) {
        return;
    }

    setCurrentUserData(widget->userData());

    setCurrentUsersWidgetPage(UsersWidgetPage::ChangeUserButton);

    m_passwordLineEdit->setFocus();
}

void LoginWidget::changeSelectingUserShortcutActivated()
{
    switch (currentAuthPage()) {
    case AuthPage::ShortCutsEnterUsername:
        setCurrentAuthPage(AuthPage::TextEnterUsername);
        break;
    case AuthPage::TextEnterUsername:
        setCurrentAuthPage(AuthPage::ShortCutsEnterUsername);
        break;
    default:
        break;
    }
}

void LoginWidget::escapeKeyShortcutActivated()
{
    if (currentUsersWidgetPage() != UsersWidgetPage::ChangeUserButton) {
        setCurrentUsersWidgetPage(UsersWidgetPage::ChangeUserButton);
    }

    if (currentCommonWidgetPage() != CommonWidgetPage::Login) {
        setCurrentCommonWidgetPage(CommonWidgetPage::Login);
    }
}

bool LoginWidget::gestureEvent(QGestureEvent* event)
{
    if (const auto swipe = event->gesture(Qt::SwipeGesture)) {
        if (const auto gesture = static_cast<QSwipeGesture*>(swipe)) {
            swipeTriggered(gesture, event->widget());
        }
    }

    return true;
}

void LoginWidget::swipeTriggered(QSwipeGesture* event, QWidget* widget)
{
    if (widget == m_allUsersTableWidget && event->state() == Qt::GestureFinished) {
        switch (event->horizontalDirection()) {
        case QSwipeGesture::Left:
            setNextUsersPage();
            break;
        case QSwipeGesture::Right:
            setPreviosUsersPage();
            break;
        default:
            break;
        }
    }
}

void LoginWidget::filterChanged(const QString &filter)
{
    UsersData filtredUsers;

    for (auto user : m_usersData) {
        if (user.username.contains(filter)) {
            filtredUsers.append(user);
        }
    }

    setAllListUsers(filtredUsers);
}

const QList<UsersData> LoginWidget::splitUsersIntoPages(const UsersData &users) const
{
    const auto usersCount = users.count();
    const auto maxCellsCount = m_allUsersRows * m_allUsersColumns;

    QList<UsersData> userPagesData;
    int leftSplitIndex = 0;
    while (true) {
        int itemCount = maxCellsCount;
        bool needBreak = false;
        if (leftSplitIndex + itemCount >= usersCount) {
            itemCount = usersCount - leftSplitIndex;
            needBreak = true;
        }
        userPagesData << users.mid(leftSplitIndex, itemCount);
        if (needBreak) {
            break;
        }
        leftSplitIndex += maxCellsCount;
    }

    return userPagesData;
}

void LoginWidget::setShortListUsers(const UsersData& users)
{
    m_shortUsersTableWidget->clear();

    m_shortUsersTableWidget->setColumnCount(std::min(m_shortUsersColumns, users.count()));

    int column = 0;
    foreach (const auto& user, users) {
        if (column >= m_shortUsersColumns) {
            break;
        }

        const auto widget = new UserCellWidget(user);
        widget->load(ThemeWidgetCreator::create(m_verticalUserWidgetFile));

        m_shortUsersTableWidget->setCellWidget(0, column++, widget);
    }
    m_shortUsersTableWidget->resizeColumnsToContents();
}

void LoginWidget::setAllListUsers(const UsersData& users)
{
    m_userPagesData = splitUsersIntoPages(users);

    const auto pagesCount = m_userPagesData.count();

    m_userPagesTableWidget->clear();
    m_userPagesTableWidget->setColumnCount(pagesCount);
    const auto tableWidgetCellWidth = m_userPagesTableWidget->horizontalHeader()->defaultSectionSize();
    m_userPagesTableWidget->setMaximumWidth(pagesCount * tableWidgetCellWidth);

    for (int column = 0; column < pagesCount; column++) {
        const auto widget = new QLabel();
        widget->setAlignment(Qt::AlignCenter);
        m_userPagesTableWidget->setCellWidget(0, column, widget);
    }

    setUsersPage(0);
}

void LoginWidget::setUserLists(const UsersData &users)
{
    if (users.count() <= m_shortUsersColumns) {
        setShortListUsers(users);
        setAllListUsers(UsersData());
    } else {
        setShortListUsers(UsersData());
        setAllListUsers(users);
    }
}

void LoginWidget::setUsersPage(const int index)
{
    m_allUsersTableWidget->clear();

    int row = 0;
    int column = 0;

    m_allUsersTableWidget->setRowCount(1);
    m_allUsersTableWidget->setColumnCount(m_allUsersColumns);

    foreach (const auto& user, m_userPagesData.value(index)) {
        if (column >= m_allUsersColumns) {
            column = 0;
            if (++row >= m_allUsersRows) {
                break;
            }
            m_allUsersTableWidget->insertRow(row);
        }

        const auto widget = new UserCellWidget(user);
        widget->load(ThemeWidgetCreator::create(m_horizontalUserWidgetFile));

        m_allUsersTableWidget->setCellWidget(row, column++, widget);
    }

    const auto pagesCount = m_userPagesTableWidget->columnCount();
    const auto tableWidgetCellHeigth = m_userPagesTableWidget->verticalHeader()->defaultSectionSize();
    for (int column = 0; column < pagesCount; column++) {
        const auto widget = qobject_cast<QLabel*>(m_userPagesTableWidget->cellWidget(0, column));
        if (!widget) {
            continue;
        }
        QIcon icon;
        if (column == index) {
            icon = m_whiteCircleIcon;
            m_currentUsersPage = column;
        } else {
            icon = m_emptyWhiteCircleIcon;
        }
        widget->setPixmap(icon.pixmap(tableWidgetCellHeigth-4, tableWidgetCellHeigth-4));
        m_userPagesTableWidget->setCellWidget(0, column, widget);
    }

    resizeAllUsersWidgets();
}

void LoginWidget::setNextUsersPage()
{
    if (m_currentUsersPage < m_userPagesTableWidget->columnCount() - 1) {
        setUsersPage(m_currentUsersPage + 1);
    }
}

void LoginWidget::setPreviosUsersPage()
{
    if (m_currentUsersPage > 0) {
        setUsersPage(m_currentUsersPage - 1);
    }
}

void LoginWidget::setCurrentUserData(const UserData& data)
{
    const auto labelSize = Utils::caclulateRealPaintDeviceSize(m_userIconLabel);

    m_userIconLabel->setPixmap(Utils::circlePixmapFromIcon(data.icon, labelSize));
    m_usernameLabel->setText(data.username);

    if (data.hostname.isEmpty()) {
        m_userHostnameLabel->setText(data.hostname);
    }
    else {
        m_userHostnameLabel->setText(tr("Welcome to ") + data.hostname);
    }

    m_passwordLineEdit->clear();

    if (data.isTokenUser) {
        m_passwordLineEdit->setPlaceholderText(tr("PIN"));
    }
    else {
        m_passwordLineEdit->setPlaceholderText(tr("Password"));
    }

    m_usernameLineEdit->setText(data.username);
    m_passwordUsernameLineEdit->clear();

    emit currentUserChanged(data.username);
}

void LoginWidget::updateUsernameCompleterValues()
{
    const auto completer = m_usernameLineEdit->completer();
    if (!completer) {
        return;
    }

    QStringList usernames;
    for (const auto& userData : m_usersData) {
        usernames << userData.username;
    }

    completer->setModel(new QStringListModel(usernames, completer));
}

void LoginWidget::resizeAllUsersWidgets()
{
    m_allUsersTableWidget->resizeColumnsToContents();

    auto maxCellWidth = m_allUsersDefaultSectionSize;
    for (int column = 0; column < m_allUsersTableWidget->columnCount(); column++) {
        maxCellWidth = std::max(m_allUsersTableWidget->horizontalHeader()->sectionSize(column), maxCellWidth);
    }

    m_allUsersTableWidget->horizontalHeader()->setDefaultSectionSize(maxCellWidth);

    QSize tableWidgetSize(maxCellWidth * m_allUsersTableWidget->columnCount(),
                          m_allUsersTableWidget->verticalHeader()->defaultSectionSize() * m_allUsersTableWidget->rowCount());
    if (parentWidget()) {
        if (parentWidget()->width() < tableWidgetSize.width()) {
            tableWidgetSize.setWidth(parentWidget()->width());
        }
    }
    m_allUsersTableWidget->setFixedSize(tableWidgetSize);
}

LoginWidget::CommonWidgetPage LoginWidget::currentCommonWidgetPage() const
{
    const auto currentPageName = m_commonStackedWidget->currentWidget()->objectName();
    if (currentPageName == "loginPage") {
        return CommonWidgetPage::Login;
    } else if (currentPageName == "allUsersPage") {
        return CommonWidgetPage::allUsers;
    } else {
        return CommonWidgetPage::UnknownPage;
    }
}

QWidget* LoginWidget::currentCommonWidget() const
{
    return m_commonStackedWidget->currentWidget();
}

void LoginWidget::setCurrentCommonWidgetPage(const LoginWidget::CommonWidgetPage page)
{
    switch (page) {
    case CommonWidgetPage::Login:
        setCommonStackedWidgetPage("loginPage");
        break;
    case CommonWidgetPage::allUsers:
        setCommonStackedWidgetPage("allUsersPage");
        break;
    default:
        return;
    }

    switch (currentAuthPage()) {
    case AuthPage::ShortCutsEnterUsername:
        m_passwordLineEdit->setFocus();
        break;
    case AuthPage::TextEnterUsername:
        QTimer::singleShot(0, [&]() {
            if (m_usernameLineEdit->text().isEmpty() || !m_focusPassword) {
                m_usernameLineEdit->setFocus();
            }
            else {
                m_passwordUsernameLineEdit->setFocus();
            }
        });
        break;
    default:
        break;
    }
}

void LoginWidget::setCommonStackedWidgetPage(const int index)
{
    m_commonStackedWidget->setCurrentIndex(index);
}

void LoginWidget::setCommonStackedWidgetPage(const QString& name)
{
    m_commonStackedWidget->setCurrentWidget(Utils::findWidget<QWidget*>(name, m_commonStackedWidget));
}

LoginWidget::UsersWidgetPage LoginWidget::currentUsersWidgetPage() const
{
    const auto currentPageName = m_usersStackedWidget->currentWidget()->objectName();
    if (currentPageName == "shortUsersPage") {
        return UsersWidgetPage::ShortCutUsers;
    } else if (currentPageName == "changeUserButtonPage") {
        return UsersWidgetPage::ChangeUserButton;
    } else {
        return UsersWidgetPage::UnknownPage;
    }
}

void LoginWidget::setCurrentUsersWidgetPage(const UsersWidgetPage page)
{
    switch (page) {
    case UsersWidgetPage::ShortCutUsers:
        setUsersStackedWidgetPage("shortUsersPage");
        break;
    case UsersWidgetPage::ChangeUserButton:
        setUsersStackedWidgetPage("changeUserButtonPage");
        break;
    default:
        break;
    }
}

void LoginWidget::setUsersStackedWidgetPage(const int index)
{
    m_usersStackedWidget->setCurrentIndex(index);
}

void LoginWidget::setUsersStackedWidgetPage(const QString& name)
{
    m_usersStackedWidget->setCurrentWidget(Utils::findWidget<QWidget*>(name, m_usersStackedWidget));
}

void LoginWidget::setLoginStackedWidgetPage(const int index)
{
    if (m_loginStackedWidget->currentIndex() != index) {
        m_loginStackedWidget->setCurrentIndex(index);
        emit authPageChanged(currentAuthPage());
    }
}

void LoginWidget::setLoginStackedWidgetPage(const QString& name)
{
    if (m_loginStackedWidget->currentWidget()->objectName() != name) {
        m_loginStackedWidget->setCurrentWidget(Utils::findWidget<QWidget*>(name, m_loginStackedWidget));
        emit authPageChanged(currentAuthPage());
    }
}

bool LoginWidget::checkUserInNoPassList(const QString &username)
{
    for (auto noPassValue : m_noPassUsers) {
        if (noPassValue.startsWith(QStringLiteral("@"))) {
            if (Utils::isUserNameInGroup(noPassValue.mid(1).toLocal8Bit().data(), username.toLocal8Bit().data())) {
                return true;
            }
        }
        else if (noPassValue == username) {
            return true;
        }
    }

    return false;
}

void LoginWidget::setFocusPassword(bool focusPassword)
{
    m_focusPassword = focusPassword;
}

void LoginWidget::removePasswordPlaceholder()
{
    m_passwordLineEdit->setPlaceholderText(QString());
}

void LoginWidget::updateHostsList(const QStringList &hosts)
{
    if (hosts.isEmpty()) {
        m_hostsComboBox->setVisible(false);
    }
    else {
        m_hostsComboBox->setVisible(true);
    }

    const QString lastIndexText = m_hostsComboBox->itemText(m_hostsComboBox->currentIndex());

    bool isNewList = hosts.count() != m_hostsComboBox->count();

    if (!isNewList) {
        for (int i = 0; i < m_hostsComboBox->count(); i++) {
            if (hosts.at(i) != m_hostsComboBox->itemText(i)) {
                isNewList = true;
                break;
            }
        }
    }

    if (isNewList) {
        m_hostsComboBox->clear();

        for (auto host : hosts) {
            m_hostsComboBox->addItem(host);
        }

        int lastHostIndex = m_hostsComboBox->findText(lastIndexText);

        if (lastHostIndex == -1) {
            m_hostsComboBox->setCurrentIndex(0);
        }
        else {
            m_hostsComboBox->setCurrentIndex(lastHostIndex);
        }
    }
}

void LoginWidget::updateUsers(const UsersData &users)
{
    m_usersData = users;

    m_usersFilterLineEdit->clear();

    setUserLists(users);
    enableTokenMessage(m_showTokenUsers && users.count() == 0);
    m_userSelectionWidget->setVisible(users.count() > 1 && (m_isUserSelectionEnabled || currentAuthPage() == AuthPage::ShortCutsEnterUsername));

    if (currentUsersWidgetPage() != UsersWidgetPage::ChangeUserButton) {
        setCurrentUsersWidgetPage(UsersWidgetPage::ChangeUserButton);
    }

    if (currentCommonWidgetPage() != CommonWidgetPage::Login) {
        setCurrentCommonWidgetPage(CommonWidgetPage::Login);
    }

    if (users.count() > 0) {
        setCurrentUserData(users.first());
    }
}

void LoginWidget::setGenericPamHeader(const QString &header, bool echo)
{
    m_usernameLabel->setText(header);

    m_passwordLineEdit->setEchoMode(echo ? QLineEdit::Normal : QLineEdit::Password);
}

void LoginWidget::setGenericPamMessage(const QString &message)
{
    m_userHostnameLabel->setText(message);
}
