#include "newpassworddialog.h"

#include <QFormLayout>
#include <QTimer>
#include <QGraphicsOpacityEffect>
#include <QPropertyAnimation>

NewPasswordDialog::NewPasswordDialog(QWidget *parent, const bool withAuth) :
    QDialog(parent),
    m_withAuth(withAuth)
{
    setWindowFlag(Qt::FramelessWindowHint);

    QFormLayout *form = new QFormLayout(this);

    m_messageLabel = new QLabel();
    showNormalMessage();
    form->addRow(m_messageLabel);
    setStyleSheet(QStringLiteral("QComboBox { font-size: 13px; background-color: #FFFFFF; border-radius: 2px; color: #000000; padding-left: 15px; } QPushButton { background-color: #5E9DCC; font-weight: bold; font-size: 13px; color: #FFFFFF; } QDialog { background-color: #808080; border-radius: 4px; }"));

    if (m_withAuth) {
        m_oldPasswordLineEdit = new QLineEdit(this);
        m_oldPasswordLineEdit->setPlaceholderText(tr("Old password"));
        m_oldPasswordLineEdit->setEchoMode(QLineEdit::Password);
        form->addRow(m_oldPasswordLineEdit);
    }
    m_newPasswordLineEdit = new QLineEdit(this);
    m_newPasswordLineEdit->setPlaceholderText(tr("New password"));
    m_newPasswordLineEdit->setEchoMode(QLineEdit::Password);
    m_repeatNewPasswordLineEdit = new QLineEdit(this);
    m_repeatNewPasswordLineEdit->setPlaceholderText(tr("Confirm new password"));
    m_repeatNewPasswordLineEdit->setEchoMode(QLineEdit::Password);

    form->addRow(m_newPasswordLineEdit);
    form->addRow(m_repeatNewPasswordLineEdit);

    m_buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel,
                               Qt::Horizontal, this);
    form->addRow(m_buttonBox);

    connect(m_buttonBox, &QDialogButtonBox::accepted, this, &NewPasswordDialog::accept);
    connect(m_buttonBox, &QDialogButtonBox::rejected, this, &NewPasswordDialog::reject);

    if (m_withAuth) {
        m_oldPasswordLineEdit->setFocus();
    }
    else {
        m_newPasswordLineEdit->setFocus();
    }
}

void NewPasswordDialog::showErrorMessage(const QString &message, const bool isPasswordChangeError)
{
    Q_UNUSED(message);

    if (isPasswordChangeError) {
        m_passwordChangeErrors++;
    }

    setEnabled(true);
    m_buttonBox->setEnabled(false);

    if (m_passwordChangeErrors >= 3) {
        m_messageLabel->setText(tr("\nThe application will be restarted."));
        m_messageLabel->setStyleSheet(QStringLiteral("QLabel { color : red; }"));
        QTimer::singleShot(2000, this, &NewPasswordDialog::handleErrorMessage);
    }
    else {
        QTimer::singleShot(1000, this, [&](){enableButtonBox();});
    }

    m_newPasswordLineEdit->clear();
    m_repeatNewPasswordLineEdit->clear();

    if (m_withAuth) {
        m_oldPasswordLineEdit->clear();
        m_oldPasswordLineEdit->setFocus();
    }
    else {
        m_newPasswordLineEdit->setFocus();
    }
}

void NewPasswordDialog::acceptDialog()
{
    QDialog::accept();
}

void NewPasswordDialog::accept()
{
    if (m_withAuth && m_oldPasswordLineEdit->text().isEmpty()) {
        showErrorMessage(tr("Enter old password"));
        m_oldPasswordLineEdit->setFocus();
        return;
    }

    if (m_newPasswordLineEdit->text().isEmpty()) {
        showErrorMessage(tr("Enter new password"));
        m_newPasswordLineEdit->setFocus();
        return;
    }

    if (m_repeatNewPasswordLineEdit->text().isEmpty()) {
        showErrorMessage(tr("Enter new password confirm"));
        m_repeatNewPasswordLineEdit->setFocus();
        return;
    }

    if (m_withAuth && m_oldPasswordLineEdit->text() == m_newPasswordLineEdit->text()) {
        showErrorMessage(tr("New and old passwords match"));
        m_newPasswordLineEdit->setFocus();
        return;
    }

    if (m_newPasswordLineEdit->text() != m_repeatNewPasswordLineEdit->text()) {
        showErrorMessage(tr("New passwords not match"));
        m_repeatNewPasswordLineEdit->setFocus();
        return;
    }

    setEnabled(false);
    emit insertionCompleted(m_withAuth ? m_oldPasswordLineEdit->text() : QString(),
                            m_newPasswordLineEdit->text(),
                            m_repeatNewPasswordLineEdit->text());
}

void NewPasswordDialog::handleErrorMessage()
{
    QGraphicsOpacityEffect *opacityEffect = new QGraphicsOpacityEffect(this);
    m_messageLabel->setGraphicsEffect(opacityEffect);

    QPropertyAnimation *messageLabelAnimation = new QPropertyAnimation(opacityEffect, "opacity");
    messageLabelAnimation->setDuration(1000);
    messageLabelAnimation->setStartValue(1.0);
    messageLabelAnimation->setEndValue(0.0);

    connect(messageLabelAnimation, &QPropertyAnimation::finished,
            this, &NewPasswordDialog::showNormalMessage);
    connect(messageLabelAnimation, &QPropertyAnimation::finished,
            this, &NewPasswordDialog::enableButtonBox);
    connect(messageLabelAnimation, &QPropertyAnimation::finished,
            opacityEffect, &QGraphicsOpacityEffect::deleteLater);
    connect(messageLabelAnimation, &QPropertyAnimation::finished,
            messageLabelAnimation, &QPropertyAnimation::deleteLater);

    messageLabelAnimation->start();
}

void NewPasswordDialog::showNormalMessage()
{
    m_messageLabel->setText(tr("Your password is expired. Please set new one."));
    m_messageLabel->setStyleSheet(QStringLiteral("QLabel { color : #FFFFFF; }"));
}

void NewPasswordDialog::enableButtonBox()
{
    m_buttonBox->setEnabled(true);
}
