#include "greeterhandler.h"

#include "kdm_greet.h"
#include "utils.h"
#include "kgreeterplugin.h"
#include "exceptions.h"

#include <memory>


GreeterHandler::GreeterHandler(const int dmSocket, QObject* parent) :
    QObject(parent),
    m_socketNotifier(dmSocket, QSocketNotifier::Read)
{
    stop();

    connect(&m_socketNotifier, &QSocketNotifier::activated,
            this, &GreeterHandler::handleSocket);
}

GreeterHandler::~GreeterHandler()
{
}

void GreeterHandler::start()
{
    m_socketNotifier.setEnabled(true);
}

void GreeterHandler::stop()
{
    m_socketNotifier.setEnabled(false);
}

const QString GreeterHandler::readLastUserSession(const QString& username)
{
    gSendInt(G_ReadDmrc);
    gSendStr(username.toUtf8());
    gRecvInt(); // ignore status code ...

    gSendInt(G_GetDmrc);
    gSendStr("Session");

    const auto sessionChar = gRecvStr();
    const auto session = QString::fromUtf8(sessionChar);
    free(sessionChar);

    return session;
}

const DpyDescriptions GreeterHandler::getDisplaySessions(const int flags)
{
    DpyDescriptions sessions;

    foreach (const auto& session, fetchSessions(flags)) {
        QString user, loc;
        decodeSession(session, user, loc);

        DpyDescription description;
        description.display = session.display;
        description.from = session.from;
        description.user = session.user;
        description.session = session.session;
        description.vt = session.vt;
        description.flags = session.flags;
        description.count = session.count;
        description.decodeLoc = loc;
        description.decodeUser = user;
        description.isCurrentSession = session.flags & isSelf;

        sessions << description;
    }

    return sessions;
}

void GreeterHandler::switchDisplay(const int vt)
{
    m_socketNotifier.setEnabled(true);

    gSet(1);
    gSendInt(G_Activate);
    gSendInt(vt);
    gSet(0);
}

void GreeterHandler::onNewPasswordInsertionCompleted(const QString &login, const QString &oldPassword, const QString &newPassword, const QString &confirmPassword)
{
    m_startedPasswordChange = true;

    m_login = login;
    m_oldPassword = oldPassword;
    m_newPassword = newPassword;
    m_confirmPassword = confirmPassword;

    handleGetTextPamResponse(m_lastGetTextPamMessage.toLocal8Bit(), m_lastGetTextPamEcho, false);
    m_lastGetTextPamMessage.clear();
}

void GreeterHandler::updateDisplaySessions()
{
    emit displaySessionsUpdated(getDisplaySessions(lstPassive | lstTTY));
}

void GreeterHandler::enableGenericMode()
{
    m_socketNotifier.setEnabled(true);

    debug("  Sending to DM %\"d\n", G_Verify);
    gSendInt(G_Verify);
    debug("  Sending to DM %\"s\n", "generic");
    gSendStr("generic");
}

void GreeterHandler::sendAnswerToPam(const QString &answer)
{
    auto code = !m_echo ?
                KGreeterPluginHandler::IsSecret :
                m_isLoginRequested ?
                    KGreeterPluginHandler::IsUser : 0;

    if (code == KGreeterPluginHandler::IsSecret) {
        debug("  Sending to DM %\"s\n", "secret");
        gSendStrNoDebug(answer.toLocal8Bit());
    }
    else {
        debug("  Sending to DM %\"s\n", answer.toLocal8Bit().data());
        gSendStr(answer.toLocal8Bit());
    }


    debug("  Sending to DM %\"d\n", code);
    gSendInt(code);
}

void GreeterHandler::handleLoginRequest(const AuthData& data)
{
    m_socketNotifier.setEnabled(true);

    debug("  Sending to DM %\"d\n", G_Verify);
    gSendInt(G_Verify);
    debug("  Sending to DM %\"s\n", "classic");
    gSendStr("classic");

    m_lastLogin = data.username;
    m_shouldSendLogin = true;
    m_lastPassword = data.password;
    m_shouldSendPassword = true;
}

void GreeterHandler::handleAutoLoginRequest()
{
    gSendInt(G_AutoLogin);
}

void GreeterHandler::handleSocket()
{
    debug("handleVerify ...\n");

    int ret = gRecvInt();

    switch (ret) {
    case V_GET_TEXT: {
        debug(" V_GET_TEXT\n");
        const auto msg = gRecvStr();
        debug("  prompt %\"s\n", msg);
        const auto echo = gRecvInt();
        m_echo = echo;
        debug("  echo = %d\n", echo);
        const auto ndelay = gRecvInt();
        debug("  ndelay = %d\n", ndelay);

        auto message = QString::fromLocal8Bit(msg);

        if (m_startedPasswordChange) {
            handleGetTextPamResponse(msg, echo, ndelay);
        }
        else {
            m_lastGetTextPamMessage = message;
            m_lastGetTextPamEcho = echo;
        }

        m_isLoginRequested = message.contains(tr("Login:"));

        emit requestReceived(message, echo);

        sendAuthDataToPam();

        free(msg);
        break;
    }
    case V_GET_BINARY: {
        debug(" V_GET_BINARY\n");
        const auto msg = gRecvArr(&ret);
        debug("  prompt %\"s\n", msg);
        debug("  %d bytes prompt\n", ret);
        const auto ndelay = gRecvInt();
        debug("  ndelay = %d\n", ndelay);
        free(msg);
        break;
    }
    case V_PUT_USER: {
        debug(" V_PUT_USER\n");
        const auto msg = gRecvStr();
        debug("  prompt %\"s\n", msg);
        free(msg);
        break;
    }
    case V_PRE_OK: {
        debug(" V_PRE_OK\n");
        break;
    }
    case V_MSG_ERR: {
        debug(" V_MSG_ERR\n");
        const auto msg = gRecvStr();
        debug("  prompt %\"s\n", msg);

        emit pamErrorReceived(QString::fromLocal8Bit(msg), m_startedPasswordChange);

        if (m_startedPasswordChange) {
            m_startedPasswordChange = false;
        }

        free(msg);
        gSendInt(0);

        break;
    }
    case V_MSG_INFO_AUTH: {
        debug(" V_MSG_INFO_AUTH\n");
        const auto msg = gRecvStr();
        debug("  prompt %\"s\n", msg);

        emit messageReceived(QString::fromLocal8Bit(msg));

        free(msg);
        gSendInt(0);

        break;
    }
    case V_MSG_INFO: {
        debug(" V_MSG_INFO\n");
        const auto msg = gRecvStr();
        debug("  display %\"s\n", msg);

        emit messageReceived(QString::fromLocal8Bit(msg));

        free(msg);
        gSendInt(0);

        return;
    }
    case V_CHTOK_AUTH: {
        debug(" V_CHTOK_AUTH\n");
        m_passwordChangeWithAuth = true;
        emit passwordChangeRequestedByPam(m_passwordChangeWithAuth);
        break;
    }
    case V_CHTOK: {
        debug(" V_CHTOK\n");
        m_passwordChangeWithAuth = false;
        emit passwordChangeRequestedByPam(m_passwordChangeWithAuth);
        break;
    }
    case V_OK: {
        debug(" V_OK\n");
        m_startedPasswordChange = false;
        emit successLogin();
        break;
    }
    case V_AUTH: {
        debug(" V_AUTH\n");
        handleAuthError();
        break;
    }
    case V_FAIL: {
        debug(" V_FAIL\n");
        handleAuthError();
        break;
    }
    default:
        logPanic(tr("Unknown V_xxx code '%1' from core").arg(ret).toStdString().data());
    }
}

void GreeterHandler::handleGetTextPamResponse(const char *response, bool echo, bool nonBlocking)
{
    Q_UNUSED(nonBlocking)

    debug(response);
    if (echo) {
        m_passwordChangeStage = 0;
        debug("stage 0");
    //} else if (!m_passwordChangeWithAuth) {
    //    m_passwordChangeStage = 1;
    //    debug("stage 1");
    } else {
        QString responseString = QString::fromLocal8Bit(response);
        if (responseString.indexOf(QRegExp("\\bpassword\\b", Qt::CaseInsensitive)) >= 0 ||
                responseString.indexOf(QRegExp(tr("\\bpassword\\b"), Qt::CaseInsensitive)) >= 0 ||
                responseString.indexOf(QRegExp("\\bPIN\\b", Qt::CaseInsensitive)) >= 0 ||
                responseString.indexOf(QRegExp(tr("\\bPIN\\b"), Qt::CaseInsensitive)) >= 0
                ) { //alex: russian too!
            if (responseString.indexOf(QRegExp("\\b(re-?(enter|type)|re?(enter|type)|again|confirm|repeat|supply)\\b",Qt::CaseInsensitive)) >= 0 ||
                    responseString.indexOf(QRegExp(tr("\\b(re-?(enter|type)|re?(enter|type)|again|confirm|repeat|supply)\\b"), //alex: russian too!
                                                   Qt::CaseInsensitive)) >= 0) {
                m_passwordChangeStage = 3;
                debug("stage 3");
            } else if (responseString.indexOf(QRegExp("\\bnew\\b", Qt::CaseInsensitive)) >= 0 ||
                       responseString.indexOf(QRegExp(tr("\\bnew\\b"), Qt::CaseInsensitive)) >= 0) { //alex: russian too!
                m_passwordChangeStage = 2;
                debug("stage 2");
            } else {
                sendMessageToPam("", KGreeterPluginHandler::IsOldPassword | KGreeterPluginHandler::IsSecret);
                return;
            }
        } else {
            emit pamErrorReceived(tr("Unrecognized prompt \"%1\"").arg(responseString), false);
            sendMessageToPam(nullptr, 0);
            m_passwordChangeStage = -1;
            debug("stage -1");
            return;
        }
    }

    sendDataToPam();
}

void GreeterHandler::sendMessageToPam(const char *text, int tag)
{
    debug("SendMessageToPam");
    debug(text);
    gSendStr(text);

    if (text) {
        gSendInt(tag);
    }
}

void GreeterHandler::sendDataToPam()
{
    debug(("Send data to pam " + QString::number(m_passwordChangeStage)).toLocal8Bit());

    switch (m_passwordChangeStage) {
    case 0:
        sendMessageToPam(m_login.toLocal8Bit(),
                                 KGreeterPluginHandler::IsUser);
        break;
    case 1:
        sendMessageToPam(m_oldPassword.toLocal8Bit(),
                                 KGreeterPluginHandler::IsPassword |
                                 KGreeterPluginHandler::IsSecret);

        break;
    case 2:
        sendMessageToPam(m_newPassword.toLocal8Bit(),
                                 KGreeterPluginHandler::IsSecret);
        break;
    default: // case 3:
        sendMessageToPam(m_confirmPassword.toLocal8Bit(),
                                 KGreeterPluginHandler::IsNewPassword |
                                 KGreeterPluginHandler::IsSecret);
        break;
    }
}

void GreeterHandler::sendAuthDataToPam()
{
    if (m_shouldSendLogin) {
        debug("  Sending to DM %\"s\n", m_lastLogin.toLocal8Bit().data());
        gSendStr(m_lastLogin.toLocal8Bit());
        m_lastLogin.clear();
        m_shouldSendLogin = false;
        debug("  Sending to DM %\"d\n", KGreeterPluginHandler::IsUser);
        gSendInt(KGreeterPluginHandler::IsUser);
    }
    else if (m_shouldSendPassword) {
        debug("  Sending to DM %\"s\n", m_lastPassword.toLocal8Bit().data());
        gSendStrNoDebug(m_lastPassword.toLocal8Bit());
        m_lastPassword.clear();
        m_shouldSendPassword = false;
        debug("  Sending to DM %\"d\n", KGreeterPluginHandler::IsPassword | KGreeterPluginHandler::IsSecret);
        gSendInt(KGreeterPluginHandler::IsPassword | KGreeterPluginHandler::IsSecret);
    }
}

void GreeterHandler::handleAuthError()
{
    m_socketNotifier.setEnabled(false);
    m_lastLogin.clear();
    m_shouldSendLogin = false;
    m_lastPassword.clear();
    m_shouldSendPassword = false;
    emit loginFailed();
}

void GreeterHandler::doShutdown(const int type, const QString &os)
{
    m_socketNotifier.setEnabled(true);

    gSet(1);
    gSendInt(G_Shutdown);
    gSendInt(type);
    gSendInt(0);
    gSendInt(0);
    gSendInt(SHUT_FORCE);
    gSendInt(0);
    gSendStr(os.toUtf8().data());
    gSet(0);
}

void GreeterHandler::setSessionType(const QString& session)
{
    m_socketNotifier.setEnabled(true);

    gSendInt(G_PutDmrc);
    gSendStr("Session");
    gSendStr(session.toUtf8());
}

void GreeterHandler::handleConsoleLoginRequest()
{
    m_socketNotifier.setEnabled(true);

    gSet(1);
    gSendInt(G_Console);
    gSet(0);
}
