#include "uigreeter.h"

#include <QScreen>
#include <QGridLayout>
#include <QSet>
#include <QApplication>
#include <QDebug>

#include "themewidgetcreator.h"
#include "kgdialog.h"
#include "kdmshutdown.h"
#include "utils.h"
#include "greeterhandler.h"
#include "shutdownhandler.h"
#include "exceptions.h"
#include "stssettings.h"
#include "managers/greetermanagers.h"
#include "widgets/basethemewidget.h"
#include "managers/sessionsmanager.h"
#include "managers/usersmanager.h"
#include "managers/shutdownmanager.h"
#include "managers/virtualkeyboardmanager.h"
#include "managers/keyboardmanager.h"
#include "managers/connectionmanager.h"
#include "managers/consoleauthmanager.h"


UiGreeter::UiGreeter(const ThemeInfo& themeInfo, const SystemInfo& systemInfo, QScreen* screen) :
    FDialog(nullptr, false),
    m_systemInfo(systemInfo),
    m_themeInfo(themeInfo),
    m_greeterHandler(new GreeterHandler(systemInfo.dmSocket)),
    m_shutdownHandler(new ShutdownHandler(m_greeterHandler)),
    m_managers(nullptr),
    m_stsSettings(new StsSettings(systemInfo.stsFile, systemInfo.dname)),
    m_wasPainted(false)
{
    connect(screen, &QScreen::geometryChanged,
            this, QOverload<const QRect&>::of(&UiGreeter::setGeometry));
    connect(&m_greeterFullTimeoutTimer, &QTimer::timeout,
            this, &UiGreeter::onGreeterFullTimeoutTimerTimeout);
    setGeometry(screen->geometry());
    setObjectName("greeterDialog");
    qApp->installEventFilter(this);
}

UiGreeter::~UiGreeter()
{
    delete m_greeterHandler;
    delete m_shutdownHandler;
    delete m_stsSettings;
    if (m_managers) {
        delete m_managers;
    }
}

void UiGreeter::load()
{
    setupManagers();

    connect(m_greeterHandler, &GreeterHandler::successLogin,
            this, &UiGreeter::performSuccessLogin);

    if (m_systemInfo.loginPlugins.isEmpty()) {
        throw CommonException(tr("Login plugins is empty"));
    }

    EnvironmentData enviroment;
    enviroment.locale = m_systemInfo.locale;
    enviroment.dataDir = m_systemInfo.dataDir;

    m_mainThemeWidget = ThemeWidgetCreator::create(m_themeInfo, m_systemInfo.loginPlugins.first(), m_managers, enviroment);

    const auto layout = new QGridLayout();
    layout->setMargin(0);
    layout->addWidget(m_mainThemeWidget);
    setLayout(layout);
}

int UiGreeter::exec()
{
    if (_greeterFullTimeout <= 0) {
        _greeterFullTimeout = 90;
    }

    m_greeterHandler->start();
    show();
    m_mainThemeWidget->exec();

    return FDialog::exec();
}

void UiGreeter::performSuccessLogin()
{
    done(ex_login);
}

void UiGreeter::restartXServer()
{
    done(EX_RESERVER_DPY);
}

void UiGreeter::paintEvent(QPaintEvent* event)
{
    FDialog::paintEvent(event);
    if (!m_wasPainted) {
        m_wasPainted = true;
        setXWindowBackgroundPixmap(backgroundPixmap());
    }
}

void UiGreeter::resizeEvent(QResizeEvent* event)
{
    FDialog::resizeEvent(event);
    if (m_wasPainted) {
        setXWindowBackgroundPixmap(backgroundPixmap());
    }
}

bool UiGreeter::eventFilter(QObject* watched, QEvent* event)
{
    if (_isReserve && isActiveWindow()) {
        switch (event->type()) {
        case QEvent::MouseButtonPress:
        case QEvent::MouseButtonRelease:
        case QEvent::MouseButtonDblClick:
        case QEvent::KeyPress:
        case QEvent::KeyRelease:
            startGreeterFullTimeoutTimer();
            break;
        default:
            break;
        }
    }

    return FDialog::eventFilter(watched, event);
}

void UiGreeter::changeEvent(QEvent* event)
{
    if (event->type() == QEvent::ActivationChange) {
        if (_isReserve && isActiveWindow()) {
            startGreeterFullTimeoutTimer();
        } else {
            m_greeterFullTimeoutTimer.stop();
        }
    }

    FDialog::changeEvent(event);
}

void UiGreeter::exitConnection()
{
    done(EX_RESERVE);
}

void UiGreeter::showConsoleDialog()
{
   const auto sessions = fetchSessions(0); // TODO
    if (KDMConfShutdown(-1, sessions, SHUT_CONSOLE, QString()).exec()) {
        m_greeterHandler->handleConsoleLoginRequest();
    }
}

void UiGreeter::reloadToXDMCPDialog()
{
    done(ex_choose);
}

void UiGreeter::startGreeterFullTimeoutTimer()
{
    m_greeterFullTimeoutTimer.start(_greeterFullTimeout * 1000);
}

void UiGreeter::onGreeterFullTimeoutTimerTimeout()
{
    done(EX_RESERVE);
}

void UiGreeter::askPamForLastUserSession(const QString &userName)
{
    emit updateLastUserSession(m_greeterHandler->readLastUserSession(userName));
}

void UiGreeter::setupManagers()
{
    const auto localAuthManager = new LocalAuthManager(m_systemInfo.autoLoginDelay, m_systemInfo.autoLoginUser);
    localAuthManager->setHostname(m_systemInfo.hostname);
    localAuthManager->setFocusPassword(m_systemInfo.focusPassword);

    connect(m_greeterHandler, &GreeterHandler::loginFailed,
            localAuthManager, &LocalAuthManager::loginFailed);
    connect(m_greeterHandler, &GreeterHandler::successLogin,
            localAuthManager, &LocalAuthManager::successLogin);
    connect(localAuthManager, &LocalAuthManager::loginRequest,
            m_greeterHandler, &GreeterHandler::handleLoginRequest);
    connect(m_greeterHandler, &GreeterHandler::passwordChangeRequestedByPam,
            localAuthManager, &LocalAuthManager::passwordChangeRequested);
    connect(localAuthManager, &LocalAuthManager::newPasswordInsertionCompleted,
            m_greeterHandler, &GreeterHandler::onNewPasswordInsertionCompleted);
    connect(localAuthManager, &LocalAuthManager::newPasswordInsertionRejected,
            this, &UiGreeter::restartXServer);
    connect(m_greeterHandler, &GreeterHandler::pamErrorReceived,
            localAuthManager, &LocalAuthManager::pamErrorReceived);
    connect(m_greeterHandler, &GreeterHandler::messageReceived,
            localAuthManager, &LocalAuthManager::messageReceived);
    connect(m_greeterHandler, &GreeterHandler::requestReceived,
            localAuthManager, &LocalAuthManager::requestReceived);
    connect(localAuthManager, &LocalAuthManager::autoLoginRequest,
            m_greeterHandler, &GreeterHandler::handleAutoLoginRequest);
    connect(localAuthManager, &LocalAuthManager::enableGenericMode,
            m_greeterHandler, &GreeterHandler::enableGenericMode);
    connect(localAuthManager, &LocalAuthManager::sendAnswerToPam,
            m_greeterHandler, &GreeterHandler::sendAnswerToPam);


    const auto sessionsDescriptions = SessionReader::readDirs(m_systemInfo.sessionsDirs);
    const auto sessionManager = new SessionsManager(sessionsDescriptions);
    connect(sessionManager, &SessionsManager::setNextSessionTypeRequest,
            m_greeterHandler, &GreeterHandler::setSessionType);

    UserInfo::Filter usersFilter;
    usersFilter.showUsers = m_systemInfo.showUsers;
    usersFilter.lowUserId = m_systemInfo.lowUserId;
    usersFilter.highUserId = m_systemInfo.highUserId;
    usersFilter.users = m_systemInfo.users;
    usersFilter.noUsers = m_systemInfo.noUsers;
    usersFilter.showRoot = m_systemInfo.showRoot;
    usersFilter.sortUsers = m_systemInfo.sortUsers;
    const auto users = UserInfo::getUserList(usersFilter);
    const auto icons = UserIcon::getUserIcons(m_systemInfo.faceDir);
    const auto usersManager = new UsersManager(users,
                                               icons,
                                               m_stsSettings,
                                               m_systemInfo.showTokenUsers,
                                               m_systemInfo.userList,
                                               m_systemInfo.defaultUser,
                                               m_systemInfo.preselUser,
                                               m_systemInfo.userCompletion,
                                               m_systemInfo.firstScreen);

    connect(usersManager, &UsersManager::askPamForLastUserSession,
            this, &UiGreeter::askPamForLastUserSession);
    connect(this, &UiGreeter::updateLastUserSession,
            usersManager, &UsersManager::updateLastUserSession);

    const auto shutdownManager = new ShutdownManager(m_systemInfo.allowShutdownNone);
    connect(shutdownManager, &ShutdownManager::shutdownRequest,
            m_shutdownHandler, &ShutdownHandler::handleShutdown);

    const auto virtualKeyboardManager = new VirtualKeyboardManager(m_systemInfo.vkbCommand, m_systemInfo.xdpi);

    const auto keyboardManager = new KeyboardManager(XKeyboard::instance());

    ConnectionManager::ConnectionData connectionData;
    connectionData.allowClose = m_systemInfo.allowClose;
    connectionData.isReserve = m_systemInfo.isReserve;
    connectionData.isLocal = m_systemInfo.isLocal;
    connectionData.isChooserEnabled = m_systemInfo.isChooserEnabled;
    const auto connectionManager = new ConnectionManager(connectionData);
    connect(connectionManager, &ConnectionManager::restartXServerRequest,
            this, &UiGreeter::restartXServer);
    connect(connectionManager, &ConnectionManager::exitConnectionRequest,
            this, &UiGreeter::exitConnection);
    connect(connectionManager, &ConnectionManager::reloadToXDMCPDialog,
            this, &UiGreeter::reloadToXDMCPDialog);
    connect(connectionManager, &ConnectionManager::switchDisplayRequest,
            m_greeterHandler, &GreeterHandler::switchDisplay);

    if (connectionManager->switchDisplayAvailable()) {
        connect(connectionManager, &ConnectionManager::displaySessionsRequest,
                m_greeterHandler, &GreeterHandler::updateDisplaySessions);
        connect(m_greeterHandler, &GreeterHandler::displaySessionsUpdated,
                connectionManager, &ConnectionManager::updateDisplaySessions);
    }

    const auto consoleAuthManager = new ConsoleAuthManager(m_systemInfo.hasConsole);
    connect(consoleAuthManager, &ConsoleAuthManager::consoleLoginRequest,
            this, &UiGreeter::showConsoleDialog);

    m_managers = new GreeterManagers({localAuthManager,
                                      sessionManager,
                                      usersManager,
                                      shutdownManager,
                                      virtualKeyboardManager,
                                      keyboardManager,
                                      connectionManager,
                                      consoleAuthManager});
}

const QPixmap UiGreeter::backgroundPixmap() const
{
    const auto screen = qApp->primaryScreen();
    return m_mainThemeWidget->backgroundPixmap().copy(0, 0, screen->size().width(), screen->size().height());
}
