#include "themewidget.h"

#include "exceptions.h"
#include "themewidgetcreator.h"
#include "utils.h"
#include "widgets/loginwidget.h"
#include "widgets/commonwidget.h"
#include "widgets/newpassworddialog.h"
#include "widgets/backgroundwidget.h"
#include "widgets/autologindialog.h"
#include "managers/virtualkeyboardmanager.h"
#include "managers/keyboardmanager.h"
#include "managers/shutdownmanager.h"
#include "managers/sessionsmanager.h"
#include "managers/connectionmanager.h"
#include "managers/consoleauthmanager.h"
#include "managers/domainmanager.h"

#include <QDebug>
#include <QGridLayout>
#include <QDir>
#include <QMessageBox>
#include <QAction>
#include <QTranslator>
#include <QSettings>
#include <kconfig.h>


ThemeWidget::ThemeWidget(QWidget *parent) :
    BaseThemeWidget(parent),
    m_themeDescription("") // TODO
{
    m_loginFailedTimer.setInterval(2000); // TODO
}

ThemeWidget::~ThemeWidget()
{
}

void ThemeWidget::load(const ThemeDescription& themeDescription, GreeterManagers* managers, const EnvironmentData& environment)
{
    m_environment = environment;
    m_themeDescription = themeDescription;
    insertManagers(managers);
    loadTranslations();

    m_sessionsManager = findManager<SessionsManager*>("sessions");

    const auto themePath = m_themeDescription.dirPath();

    m_commonWidget = new CommonWidget();
    m_commonWidget->load(ThemeWidgetCreator::create(themePath + "/common_widget.ui"));
    m_commonWidget->setAvailableSessions(m_sessionsManager->availableSessions());

    m_loginWidget = new LoginWidget();
    const auto sourceLoginWidget = ThemeWidgetCreator::create(themePath + "/central_widget.ui");
    m_loginWidget->loadGeneric(sourceLoginWidget, m_themeDescription);
    m_loginWidget->enableHostnameSelection(false);

    m_backgroundWidget = new BackgroundWidget(this);
    ThemeSettings::BackgroundData backgroundData;
    try {
        m_settings.load(themePath + "/settings.yaml");
        backgroundData = m_settings.background();
    } catch (std::exception& exception) {
        qDebug() << QString::fromStdString(exception.what());
    }
    if (backgroundData.path.defined && backgroundData.path.value == "fly-default-modern") {
        if (!backgroundData.blur.enable.defined) {
            backgroundData.blur.enable.value = true;
        }
        if (!backgroundData.blur.radius.defined) {
            backgroundData.blur.radius.value = 5.0;
        }
        if (!backgroundData.logo.enable.defined) {
            backgroundData.logo.enable.value = true;
        }
        if (!backgroundData.colorOverlay.defined) {
            backgroundData.colorOverlay.defined = true;
            backgroundData.colorOverlay.value = QColor(0, 0, 0, 102);
        }

        bool exists;
        const auto dr = Utils::screenGeometry();
        const auto path = QString::fromStdString(Utils::FindFlyWallPaper(dr.width(), dr.height(), &exists, environment.dataDir));
        m_backgroundWidget->setImage(path);
    } else if (backgroundData.path.defined) {
        m_backgroundWidget->setImage(QDir(themePath).absoluteFilePath(backgroundData.path.value));
    }
    if (backgroundData.logo.enable.value) {
        if (backgroundData.logo.path.defined) {
            m_backgroundWidget->setLogo(backgroundData.logo.path.value);
        } else {
            //from dm_greet
            const auto dr = Utils::screenGeometry();
            QString logoPath = "/usr/share/wallpapers/logo/astra_logo";
            logoPath = QString::fromStdString(Utils::FlyDefaultLogoPath(logoPath.toStdString(), dr.width(), dr.height(),std::string("white")));
            m_backgroundWidget->setLogo(QDir(themePath).absoluteFilePath(logoPath));
        }
    }
    if (backgroundData.blur.enable.value) {
        m_backgroundWidget->setBlurRadius(backgroundData.blur.radius.value);
    }
    if (backgroundData.colorOverlay.defined) {
        m_backgroundWidget->setColor(backgroundData.colorOverlay.value);
    }

    const auto layout = new QGridLayout();
    layout->setMargin(0);
    layout->addWidget(m_commonWidget);
    setLayout(layout);

    const auto centralGridLayout = findWidget<QGridLayout*>("centralGridLayout");
    centralGridLayout->addWidget(m_loginWidget);

    const auto dateTimeWidget = findWidget<QWidget*>("dateTimeWidget");
    m_dateTimeWidget.load(dateTimeWidget);
    m_dateTimeWidget.start();

    m_usersManager = findManager<UsersManager*>("users");

    m_keyboardManager = findManager<KeyboardManager*>("keyboard");
    connect(m_keyboardManager, &KeyboardManager::capsLockStateChanged,
            this, &ThemeWidget::capsLockStateChangedHandle);
    connect(m_commonWidget, &CommonWidget::changeLayoutRequest,
            m_keyboardManager, &KeyboardManager::changeLayout);
    connect(m_keyboardManager, &KeyboardManager::layoutChanged,
            m_commonWidget, &CommonWidget::changeLayout);
    m_commonWidget->changeLayout(m_keyboardManager->getCurrentLayout());
    QTimer::singleShot(1000, this, [&](){
        m_commonWidget->changeLayout(m_keyboardManager->getCurrentLayout());
    });

    m_virtualKeyboardManager = findManager<VirtualKeyboardManager*>("virtualkeyboard");
    connect(m_commonWidget, &CommonWidget::openVirtualKeyboardRequest,
            m_virtualKeyboardManager, &VirtualKeyboardManager::open);
    connect(m_virtualKeyboardManager, &VirtualKeyboardManager::errorOccurred,
            this, &ThemeWidget::virtualKeyboardErrorHandle);

    m_shutdownManager = findManager<ShutdownManager*>("shutdown");
    const auto enableShutdown = !m_shutdownManager->allowShutdownNone();
    if (enableShutdown) {
        connect(m_commonWidget, &CommonWidget::suspendRequest,
                this, &ThemeWidget::suspendToRamHandle);
        connect(m_commonWidget, &CommonWidget::rebootRequest,
                this, &ThemeWidget::rebootHandle);
        connect(m_commonWidget, &CommonWidget::shutdownRequest,
                this, &ThemeWidget::shutdownHandle);
    }
    m_commonWidget->enableShutdownWidgets(enableShutdown);

    QList<QAction*> actions;
    QList<QMenu *> menus;

    m_connectionManager = findManager<ConnectionManager*>("connection");
    const auto exitAction = m_connectionManager->exitAction();
    if (exitAction.isAvailable) {
        const auto action = new QAction(exitAction.text, this);
        action->setShortcut(QKeySequence(Qt::ALT + Qt::Key_E));
        action->setShortcutContext(Qt::ApplicationShortcut);
        connect(action, &QAction::triggered,
                this, &ThemeWidget::exitActionTriggered);
        actions << action;
    }

    if (m_connectionManager->switchDisplayAvailable()) {
        menus << createSelectSessionMenu();
    }

    if (m_connectionManager->chooserDialogAvailable()) {
        const auto action = new QAction(tr("Remote login"));
        action->setShortcut(QKeySequence(Qt::ALT + Qt::Key_T));
        action->setShortcutContext(Qt::ApplicationShortcut);
        connect(action, &QAction::triggered,
                m_connectionManager, &ConnectionManager::reloadToXDMCPDialog);
        actions << action;
    }

    m_consoleAuthManager = findManager<ConsoleAuthManager*>("consoleAuth");
    if (m_consoleAuthManager->hasConsole()) {
        const auto consoleLoginAction = new QAction(tr("Console login"), this);
        consoleLoginAction->setShortcut(QKeySequence(Qt::ALT + Qt::Key_N));
        consoleLoginAction->setShortcutContext(Qt::ApplicationShortcut);
        connect(consoleLoginAction, &QAction::triggered,
                this, &ThemeWidget::consoleLoginhandle);
        actions << consoleLoginAction;
    }

    if (enableShutdown) {
        const auto rebootAction = new QAction(tr("Reboot"));
        rebootAction->setShortcut(QKeySequence(Qt::ALT + Qt::Key_R));
        rebootAction->setShortcutContext(Qt::ApplicationShortcut);
        connect(rebootAction, &QAction::triggered,
                this, &ThemeWidget::rebootHandle);
        actions << rebootAction;

        const auto shutdownAction = new QAction(tr("Shutdown"));
        shutdownAction->setShortcut(QKeySequence(Qt::ALT + Qt::Key_S));
        shutdownAction->setShortcutContext(Qt::ApplicationShortcut);
        connect(shutdownAction, &QAction::triggered,
                this, &ThemeWidget::shutdownHandle);
        actions << shutdownAction;

        const auto sleepAction = new QAction(tr("Suspend"));
        sleepAction->setShortcut(QKeySequence(Qt::ALT + Qt::Key_D));
        sleepAction->setShortcutContext(Qt::ApplicationShortcut);
        connect(sleepAction, &QAction::triggered,
                this, &ThemeWidget::sleepHandle);
        actions << sleepAction;

        const auto suspendToRamAction = new QAction(tr("Hibernation"));
        suspendToRamAction->setShortcut(QKeySequence(Qt::ALT + Qt::Key_H));
        suspendToRamAction->setShortcutContext(Qt::ApplicationShortcut);
        connect(suspendToRamAction, &QAction::triggered,
                this, &ThemeWidget::suspendToRamHandle);
        actions << suspendToRamAction;
    }

    m_commonWidget->setActions(actions, menus);

    m_localAuthManager = findManager<LocalAuthManager*>("localAuth");

    m_loginWidget->setFocusPassword(m_localAuthManager->focusPassword());
    connect(m_localAuthManager, &LocalAuthManager::loginFailed,
            this, &ThemeWidget::performLoginFailed);
    connect(m_localAuthManager, &LocalAuthManager::successLogin,
            this, &ThemeWidget::performSuccessLogin);
    connect(m_loginWidget, &LoginWidget::submitAnswerToPam,
            m_localAuthManager, &LocalAuthManager::sendAnswerToPam);

    connect(&m_loginFailedTimer, &QTimer::timeout,
            this, &ThemeWidget::loginFailedTimerTimeout);

    connect(m_localAuthManager, &LocalAuthManager::pamErrorReceived,
            this, &ThemeWidget::showErrorMessage);

//    connect(m_localAuthManager, &LocalAuthManager::messageReceived,
//            this, &ThemeWidget::showMessage);
    connect(m_localAuthManager, &LocalAuthManager::messageReceived,
            m_loginWidget, &LoginWidget::setGenericPamMessage);
    connect(m_localAuthManager, &LocalAuthManager::requestReceived,
            m_loginWidget, &LoginWidget::setGenericPamHeader);

    m_loginWidget->removePasswordPlaceholder();

    emit m_localAuthManager->enableGenericMode();
}

void ThemeWidget::exec()
{
    if (m_localAuthManager->autoLoginEnabled()) {
        m_loginWidget->setCurrentUser(m_localAuthManager->autoLoginUser());
        m_loginWidget->setCurrentAuthPage(LoginWidget::AuthPage::TextEnterUsername);
        m_commonWidget->setEnabled(false);
        AutoLoginDialog dialog(m_localAuthManager->autoLoginUser(), m_localAuthManager->autoLoginDelay(), this);
        if (dialog.exec() == QDialog::Accepted) {
            connect(m_localAuthManager, &LocalAuthManager::loginFailed, [&]() {
                m_commonWidget->setEnabled(true);
            });
            m_localAuthManager->performAutoLoginRequest();
        } else {
            m_commonWidget->setEnabled(true);
        }
    }
}

const QPixmap ThemeWidget::backgroundPixmap() const
{
    return m_backgroundWidget->grab(QRect(QPoint(0, 0), size()));
}

void ThemeWidget::performLoginRequest(const AuthData& data)
{
    m_loginWidget->enableLoginWidgets(false);
    m_commonWidget->resetFocus(); // TODO

    const auto selectedSessionType = m_commonWidget->selectedSessionType();
    if (!selectedSessionType.isEmpty()) {
        m_sessionsManager->setNextSessionType(selectedSessionType);
    }

    m_localAuthManager->performLoginRequest(data);
}

void ThemeWidget::performLoginFailed()
{
    m_commonWidget->enableLoginFailedWidget();
    m_loginFailedTimer.start();
}

void ThemeWidget::performSuccessLogin()
{
    m_usersManager->savePrevUsername(m_loginWidget->currentUsername());
}

void ThemeWidget::loginFailedTimerTimeout()
{
    m_loginFailedTimer.stop();
    m_commonWidget->enableLoginFailedWidget(false);
    m_loginWidget->clearPassword();
    m_loginWidget->enableLoginWidgets();
    m_loginWidget->setFocusOnPassword();
    emit m_localAuthManager->sendAnswerToPam(QStringLiteral("0"));
}

void ThemeWidget::virtualKeyboardErrorHandle(const QString& error)
{
    QMessageBox::critical(this,
                          tr("Virtual keyboard error"),
                          error);
}

void ThemeWidget::capsLockStateChangedHandle(const bool lock)
{
    m_loginWidget->enableCapsLock(lock);
}

void ThemeWidget::rebootHandle()
{
    m_commonWidget->showActionsMenu(false);
    m_shutdownManager->doShutdown(ShutdownAction::Reboot);
}

void ThemeWidget::shutdownHandle()
{
    m_commonWidget->showActionsMenu(false);
    m_shutdownManager->doShutdown(ShutdownAction::Shutdown);
}

void ThemeWidget::sleepHandle()
{
    m_commonWidget->showActionsMenu(false);
    m_shutdownManager->doShutdown(ShutdownAction::Sleep);
}

void ThemeWidget::suspendToRamHandle()
{
    m_commonWidget->showActionsMenu(false);
    m_shutdownManager->doShutdown(ShutdownAction::SuspendToRam);
}

void ThemeWidget::consoleLoginhandle()
{
    m_commonWidget->showActionsMenu(false);
    m_consoleAuthManager->performConsoleLogin();
}

void ThemeWidget::exitActionTriggered()
{
    m_commonWidget->showActionsMenu(false);
    m_connectionManager->performExitAction();
}

void ThemeWidget::changeAuthPageActionTriggered()
{
    m_loginWidget->setCurrentAuthPage(static_cast<LoginWidget::AuthPage>(m_changeAuthPageAction->data().toInt()));
}

void ThemeWidget::authPageChanged()
{
    QString text;
    auto pageToChange = LoginWidget::AuthPage::UnknownPage;

    switch (m_loginWidget->currentAuthPage()) {
    case LoginWidget::AuthPage::ShortCutsEnterUsername:
        text = tr("Enter username");
        pageToChange = LoginWidget::AuthPage::TextEnterUsername;
        break;
    case LoginWidget::AuthPage::TextEnterUsername:
        text = tr("Choose user");
        pageToChange = LoginWidget::AuthPage::ShortCutsEnterUsername;
        break;
    default:
        break;
    }

    if (text.isEmpty()) {
        m_changeAuthPageAction->setVisible(false);
    } else {
        m_changeAuthPageAction->setVisible(true);
    }

    m_changeAuthPageAction->setText(text);
    m_changeAuthPageAction->setData(static_cast<int>(pageToChange));
}

void ThemeWidget::openNewPasswordDialog(const bool withAuth)
{
    NewPasswordDialog *newPasswordDialog = new NewPasswordDialog(this, withAuth);

    connect(newPasswordDialog, &NewPasswordDialog::insertionCompleted,
            this, &ThemeWidget::onNewPasswordInsertionCompleted);
    connect(this, &ThemeWidget::newPasswordInsertionCompleted,
            m_localAuthManager, &LocalAuthManager::newPasswordInsertionCompleted);
    connect(newPasswordDialog, &NewPasswordDialog::rejected,
            m_localAuthManager, &LocalAuthManager::newPasswordInsertionRejected);
    connect(m_localAuthManager, &LocalAuthManager::successLogin,
            newPasswordDialog, &NewPasswordDialog::acceptDialog);
    connect(m_localAuthManager, &LocalAuthManager::pamErrorReceived,
            newPasswordDialog, &NewPasswordDialog::showErrorMessage);
    connect(newPasswordDialog, &NewPasswordDialog::rejected,
            newPasswordDialog, &NewPasswordDialog::deleteLater);

    m_isNewPasswordDialogOpened = true;
    newPasswordDialog->show();
    QApplication::setActiveWindow(newPasswordDialog);
}

void ThemeWidget::showErrorMessage(const QString &message)
{
    QMessageBox::critical(this, "Error", message);
}

void ThemeWidget::showMessage(const QString &message)
{
    if (!message.isEmpty()) {
        QMessageBox::warning(this, "Warning", message);
    }
}

void ThemeWidget::loadTranslations()
{
    const auto locale = m_environment.locale;

    if (locale.isNull() || locale.isEmpty()) {
        return;
    }

    const auto themeDir = QFileInfo(m_themeDescription.path()).dir();
    const auto translationsFiles = themeDir.entryInfoList(QStringList() << QString("*_%1.qm").arg(locale),
                                                          QDir::Files | QDir::Readable);
    foreach (const auto file, translationsFiles) {
        const auto translator = new QTranslator();
        if (translator->load(file.absoluteFilePath())) {
            qApp->installTranslator(translator);
        } else {
            delete translator;
        }
    }
}

void ThemeWidget::resizeEvent(QResizeEvent* event)
{
    BaseThemeWidget::resizeEvent(event);
    m_backgroundWidget->resize(size());
}

void ThemeWidget::checkNoPassSettings()
{
    if (!m_connectionManager->switchDisplayAvailable()) {
        return;
    }

    auto entryMap = KConfig(QStringLiteral("/etc/X11/fly-dm/fly-dmrc")).entryMap(QStringLiteral("X-:*-Core"));

    bool isNoPassEnable = entryMap.value(QStringLiteral("NoPassEnable")).contains(QStringLiteral("true"));
    QStringList noPassUsersList = entryMap.value(QStringLiteral("NoPassUsers")).split(QLatin1Char(','), QString::SkipEmptyParts);

    if (isNoPassEnable && !noPassUsersList.isEmpty()) {
        m_loginWidget->setNoPass(isNoPassEnable, noPassUsersList);
    }
}

QMenu *ThemeWidget::createSelectSessionMenu()
{
    QMenu *switchSessionMenu = new QMenu(tr("Switch session"), m_commonWidget);

    switchSessionMenu->setWindowFlags((windowFlags() ^ Qt::Window) | Qt::Widget);
    switchSessionMenu->hide();

    switchSessionMenu->menuAction()->setShortcut(QKeySequence(Qt::ALT + Qt::Key_I));
    switchSessionMenu->menuAction()->setShortcutContext(Qt::ApplicationShortcut);

    connect(switchSessionMenu->menuAction(), &QAction::triggered,
            this, [&, switchSessionMenu](){
        if (switchSessionMenu->isHidden()) {
            auto actionsMenu = m_commonWidget->m_actionsToolButton->menu();

            if (actionsMenu->isHidden()) {
                QTimer::singleShot(0, this, [actionsMenu](){
                    actionsMenu->hide();
                });

                m_commonWidget->showActionsMenu();
            }

            int switchSessionMenuX = actionsMenu->x() + actionsMenu->width();
            switchSessionMenu->popup(QPoint(switchSessionMenuX, actionsMenu->y()));

            QTimer::singleShot(0, this, [&]() {
                m_commonWidget->showActionsMenu();
            });
        }
        else {
            switchSessionMenu->hide();
        }
    });

    connect(switchSessionMenu, &QMenu::triggered,
            [&](QAction *action){
        if (action->isChecked()) {
            m_connectionManager->switchDisplay(action->data().toInt());
        }
    });
    connect(switchSessionMenu, &QMenu::aboutToShow,
            m_connectionManager, &ConnectionManager::displaySessionsRequest);
    connect(m_connectionManager, &ConnectionManager::updateDisplaySessions,
            [this, switchSessionMenu](const DpyDescriptions &sessions){
        switchSessionMenu->clear();

        for (const auto &session : sessions) {
            QAction *action = switchSessionMenu->addAction(tr( "%1 (%2)" ).arg(session.decodeUser).arg(session.decodeLoc));
            action->setData(session.vt ? session.vt : -1);
            action->setCheckable(true);

            if (!session.vt) {
                action->setEnabled(false);
            }

            if (session.isCurrentSession) {
                action->setChecked(true);
            }
        }

        emit dpyMenuUpdatingCompleted();
    });


    return switchSessionMenu;
}

void ThemeWidget::onNewPasswordInsertionCompleted(const QString &oldPassword,
                                                  const QString &newPassword,
                                                  const QString &confirmPassword)
{
    emit newPasswordInsertionCompleted(m_loginWidget->currentUsername(),
                                       oldPassword.isEmpty() ? m_loginWidget->currentPassword() : oldPassword,
                                       newPassword,
                                       confirmPassword);
}

template<typename T>
T ThemeWidget::findWidget(const QString& name) const
{
    return Utils::findWidget<T>(name, this);
}

template<typename T>
T ThemeWidget::findManager(const QString& id) const
{
    const auto point = m_managers->get(id);
    if (!point) {
        throw CommonException(tr("Greeter manager '%1' not found").arg(id));
    }

    const auto manager = dynamic_cast<T>(point);
    if (!manager) {
        throw CommonException(tr("Menager '%1' convertation error").arg(id));
    }

    return manager;
}

PluginThemeLoader::PluginThemeLoader(QObject *parent) :
    AbstractPluginThemeLoader(parent)
{

}

PluginThemeLoader::~PluginThemeLoader()
{

}

BaseThemeWidget *PluginThemeLoader::createTheme() const
{
    return new ThemeWidget();
}
